# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import itertools
from collections import Callable
from typing import Any, Tuple, Dict

import numpy as np

from model_compression_toolkit.core.common.constants import MIN_THRESHOLD, DEFAULT_TOL, DEFAULT_DEC_FACTOR, \
    SYMMETRIC_TENSOR_PER_CHANNEL_N_INTERVALS, SYMMETRIC_TENSOR_PER_CHANNEL_N_ITER, SYMMETRIC_TENSOR_DEC_FREQ, \
    SYMMETRIC_TENSOR_PER_CHANNEL_DEC_FREQ, SYMMETRIC_TENSOR_N_INTERVALS, SYMMETRIC_TENSOR_N_ITER, \
    UNIFORM_TENSOR_PER_CHANNEL_N_ITER, UNIFORM_TENSOR_N_ITER, SYMMETRIC_HISTOGRAM_DEC_FREQ, SYMMETRIC_HISTOGRAM_N_ITER, \
    SYMMETRIC_HISTOGRAM_N_INTERVALS, UNIFORM_HISTOGRAM_N_ITER, BOTTOM_FACTOR, UPPER_FACTOR, UNIFORM_TENSOR_N_SAMPLES, \
    UNIFORM_HISTOGRAM_N_SAMPLES, DEC_RANGE_UPPER, DEC_RANGE_BOTTOM
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import quantize_tensor, \
    reshape_tensor_for_per_channel_search, uniform_quantize_tensor, get_output_shape
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import max_power_of_two, \
    get_tensor_max

def qparams_selection_tensor_search(error_function: Callable,
                                    tensor_data: np.ndarray,
                                    n_bits: int,
                                    per_channel: bool = False,
                                    channel_axis: int = 1,
                                    n_iter: int = 10,
                                    min_threshold=MIN_THRESHOLD) -> Any:
    """
    Search for an optimal threshold to quantize a tensor.
    The search_methods starts with the constrained no-clipping threshold the tensor has, and continues with
    n_iter another smaller constrained thresholds. For each candidate threshold, an error is computed
    based on the passed error function, and the threshold which yields the minimal error is selected
    and returned.

    Args:
        error_function: Function to compute the error between the original and quantized tensors.
        tensor_data: Numpy array with tensor's content.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the tensor should be quantized per-channel or per-tensor.
        channel_axis: Index of output channels dimension.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Optimal constrained threshold to quantize the tensor.

    """

    signed = np.any(tensor_data < 0)  # check if tensor is singed
    output_shape = get_output_shape(tensor_data.shape, channel_axis)

    # First threshold to check is the constrained threshold based on the tensor's maximal value.
    tensor_max = get_tensor_max(np.abs(tensor_data), per_channel, channel_axis)
    threshold = 2 * max_power_of_two(tensor_max, min_threshold)

    # If the threshold is computed per-channel, we rearrange the tensor such that each sub-tensor
    # is flattened, and we iterate over each one of them when searching for the threshold.
    if per_channel:
        tensor_data_r = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)

    error_list = []  # init an empty error list
    # On each iteration a new constrained threshold which equal to half of the previous tested threshold
    # is used for quantizing the tensor and computing the error. The error is appended to an error list, which
    # eventually used to select the threshold with the minimal error.
    for i in range(n_iter):
        if per_channel:
            threshold_hat = (threshold / (2 ** i)).reshape([-1, 1])
            qt = quantize_tensor(tensor_data_r, threshold_hat, n_bits, signed)
            per_channel_error = _error_function_wrapper(error_function, tensor_data_r, qt, threshold_hat)

            error_list.append(per_channel_error)
        else:  # quantize per-tensor
            qt = quantize_tensor(tensor_data, threshold / (2 ** i), n_bits, signed)
            error = error_function(qt, tensor_data, threshold=threshold / (2 ** i))
            error_list.append(error)

    # Take the index of the minimal error, and use it compute the threshold which yielded it.
    i = np.argmin(np.stack(error_list, axis=-1), axis=-1)

    return np.maximum(np.reshape(threshold.flatten() / np.power(2, i), output_shape), min_threshold)


def qparams_selection_histogram_search(error_function: Callable,
                                       bins: np.ndarray,
                                       counts: np.ndarray,
                                       n_bits: int,
                                       constrained: bool = True,
                                       n_iter: int = 10,
                                       min_threshold: float = MIN_THRESHOLD):
    """
    Search for an optimal threshold to quantize a histogram of collected float values.
    The search_methods starts with the constrained no-clipping threshold by the bins' maximal value, and continues with
    n_iter another smaller constrained thresholds. For each candidate threshold, an error is computed
    based on the passed error function, and the threshold which yields the minimal error is selected
    and returned.

    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        bins: Bins of the histogram to search_methods for an optimal threshold.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        constrained: Whether the threshold should be constrained or not.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Optimal constrained threshold to quantize the tensor.

    """

    signed = np.any(bins < 0)  # Whether histogram contains negative values or not.
    tensor_data = np.abs(bins)
    tensor_max = np.max(tensor_data)
    if not constrained:
        return tensor_max
    threshold = (1 + int(constrained)) * max_power_of_two(tensor_max, min_threshold=min_threshold)

    # Init a list of thresholds.
    error_list = []
    threshold_list = threshold / np.power(2, np.linspace(0, n_iter - 1, n_iter))

    # On each iteration a new constrained threshold which equal to half of the previous tested threshold
    # is used for quantizing the histogram and computing the error. The error is appended to an error list, which
    # eventually used to select the threshold with the minimal error.
    for threshold in threshold_list:
        q_bins = quantize_tensor(bins, threshold, n_bits, signed)  # compute the quantized values of the bins.
        error = qparams_selection_histogram_search_error_function(error_function, bins, q_bins, counts,
                                                                  threshold=threshold)
        error_list.append(error)

    # Return the threshold with the minimal error.
    return np.maximum(threshold_list[np.argmin(error_list)], min_threshold)


def qparams_symmetric_iterative_minimization(x0: np.ndarray,
                                             x: np.ndarray,
                                             loss_fn: Callable,
                                             n_bits: int,
                                             signed: bool = True,
                                             n_intervals: int = SYMMETRIC_TENSOR_N_INTERVALS,
                                             n_iter: int = SYMMETRIC_TENSOR_N_ITER,
                                             alpha: float = BOTTOM_FACTOR,
                                             beta: float = UPPER_FACTOR,
                                             dec_factor: Tuple = DEFAULT_DEC_FACTOR,
                                             dec_freq: int = SYMMETRIC_TENSOR_DEC_FREQ,
                                             tolerance: float = DEFAULT_TOL,
                                             per_channel=False) -> Dict[str, np.ndarray]:
    """
    Search for an optimal threshold to for symmetric tensor quantization.
    The search starts with the no-clipping threshold the tensor has, and continues with
    n_iter of iterative search. In each iteration, a set of n_intervals threshold is created by evenly-spacing the
    range constructed from the previous obtained threshold multiplied by the alph and beta factors (for lower and upper
    limits, respectively).
    In addition, each dec_freq iterations, the factors are multiplied by dec_factor in order to make the search range
    narrower.

        Args:
        x0: Initial threshold.
        x: Numpy array with tensor's content.
        loss_fn: Function to compute the error between the original and quantized tensors.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether quantization range is signed or not.
        n_intervals: Number of locations to examine each iteration from the given range.
        n_iter: Number of searching iterations.
        alpha: Factor for creating the lower limit of the search range.
        beta: Factor for creating the upper limit of the search range.
        dec_factor: Factor for decreasing the multiplication factors, to get narrower search range.
        dec_freq: Frequency for decreasing the multiplication factors.
        tolerance: If the improvement between iterations is smaller than tolerance, then early stop.
        per_channel: Whether quantization is done per-channel or per-tensor.

    Returns:
        Dictionary with optimized threshold for symmetric tensor quantization (best obtained during the search),
        and its matching loss value.

    """
    range_scale = np.array([alpha, beta])
    curr_threshold = x0

    if per_channel:
        # wrapping loss function with per-channel wrapper for vectorized per-channel computation
        # Note: x should be already reshaped tensor for per-channel search
        curr_threshold = curr_threshold.reshape([-1, 1])
        loss = _error_function_wrapper(loss_fn, x, quantize_tensor(x, curr_threshold, n_bits, signed), curr_threshold).reshape([-1, 1])
    else:
        loss = loss_fn(x, quantize_tensor(x, curr_threshold, n_bits, signed), curr_threshold)

    best = {"param": curr_threshold, "loss": loss}

    for n in range(n_iter):
        prev_best_loss = best['loss']
        new_range_bounds = curr_threshold * range_scale

        curr_res = search_fixed_range_intervals(new_range_bounds, x, loss_fn, n_bits, signed, n_intervals, per_channel)
        curr_threshold = curr_res['param']
        curr_loss = curr_res['loss']

        imp_losses = curr_loss < prev_best_loss
        best = {"param": curr_threshold * imp_losses + best['param'] * np.logical_not(imp_losses),
                "loss": curr_loss * imp_losses + prev_best_loss * np.logical_not(imp_losses)}

        iters_loss_diff_avg = np.mean(prev_best_loss - curr_loss)
        if 0 < iters_loss_diff_avg < tolerance:
            # improvement in last step is very small on average, therefore - finishing the search
            break

        # increase scaler to make range bounds narrower in next iteration
        if n % dec_freq == 0:
            range_scale *= dec_factor
            # prevent min bound from exceeding max bound
            range_scale = np.array([min(range_scale[0], DEC_RANGE_BOTTOM), max(range_scale[1], DEC_RANGE_UPPER)])

    return best


def iterative_uniform_dynamic_range_search(x0: np.ndarray,
                                           x: np.ndarray,
                                           scalers: np.ndarray,
                                           loss_fn: Callable,
                                           n_bits: int,
                                           n_iter: int = UNIFORM_TENSOR_N_ITER,
                                           tolerance: float = DEFAULT_TOL,
                                           per_channel: bool = False) -> Dict[str, np.ndarray]:
    """
    Search for an optimal quantization range for uniform tensor quantization.
    The search starts with the no-clipping range the tensor has, and continues with
    n_iter of iterative search. In each iteration, a set of range candidates is created by multiplying each
    scaler (from the scalers set) with the current base range (the last obtained range).

        Args:
        x0: Initial base range.
        x: Numpy array with tensor's content.
        loss_fn: Function to compute the error between the original and quantized tensors.
        scalers: A set of multiplication factors, to create a set of quantization range candidates at each iteration.
        n_bits: Number of bits to quantize the tensor.
        n_iter: Number of searching iterations.
        tolerance: If the improvement between iterations is smaller than tolerance, then early stop.
        per_channel: Whether quantization is done per-channel or per-tensor.

    Returns:
        Dictionary with optimized quantization range for uniform tensor quantization (best obtained during the search),
        and its matching loss value.

    """
    curr_range_bounds = x0

    if per_channel:
        # wrapping loss function with per-channel wrapper for vectorized per-channel computation
        # Note: x should be already reshaped tensor for per-channel search and x0 is a tensor or ranges
        # of shape (num_channels, 2)
        loss = _error_function_wrapper(loss_fn, x, uniform_quantize_tensor(x,
                                                                           curr_range_bounds[:, 0].reshape([-1, 1]),
                                                                           curr_range_bounds[:, 1].reshape([-1, 1]),
                                                                           n_bits),
                                       curr_range_bounds).reshape([-1, 1])
    else:
        loss = loss_fn(x, uniform_quantize_tensor(x, curr_range_bounds[0], curr_range_bounds[1], n_bits),
                       curr_range_bounds)

    best = {"param": curr_range_bounds, "loss": loss}

    for n in range(n_iter):
        prev_best_loss = best['loss']
        curr_res = search_dynamic_range(base_range=curr_range_bounds, scalers=scalers, x=x, loss_fn=loss_fn,
                                        n_bits=n_bits, per_channel=per_channel)
        curr_range_bounds = curr_res['param']
        curr_loss = curr_res['loss']

        imp_losses = curr_loss < prev_best_loss
        best = {"param": curr_range_bounds * imp_losses + best['param'] * np.logical_not(imp_losses),
                "loss": curr_loss * imp_losses + prev_best_loss * np.logical_not(imp_losses)}

        iters_loss_diff_avg = np.mean(prev_best_loss - curr_loss)
        if 0 < iters_loss_diff_avg < tolerance:
            # improvement in last step is very small, therefore - finishing the search
            break

    return best


def search_fixed_range_intervals(range_bounds: np.ndarray,
                                 x: np.ndarray,
                                 loss_fn: Callable,
                                 n_bits: int,
                                 signed: bool = True,
                                 n_intervals: int = 100,
                                 per_channel: bool = False) -> Dict[str, np.ndarray]:
    """
    Searches in a set of n_intervals thresholds, taken from evenly-space intervales from the constructed range.

    Args:
        range_bounds: A range for creating a set of evenly-spaced threshold candidates.
        x: Numpy array with tensor's content.
        loss_fn: Function to compute the error between the original and quantized tensors.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether quantization range is signed or not.
        n_intervals: Number of locations to examine each iteration from the given range.
        per_channel: Whether the search is done per-channel or per-tensor.

    Returns: Dictionary with best obtained threshold and the threshold's matching loss.

    """
    if per_channel:
        # search per-channel
        intervals = np.linspace(start=range_bounds[:, 0], stop=range_bounds[:, 1], num=n_intervals, dtype=float)
        # just the first interval values
        first_interval_thresholds = intervals[0, :].reshape([-1, 1])
        best = {"param": first_interval_thresholds,
                "loss": _error_function_wrapper(loss_fn, x,
                                                quantize_tensor(x, first_interval_thresholds, n_bits, signed),
                                                first_interval_thresholds).reshape([-1, 1])}
        for interval in range(1, n_intervals):
            interval_thresholds = intervals[interval, :].reshape([-1, 1])  # takes the i'th threshold for each channel
            interval_losses = _error_function_wrapper(loss_fn, x, quantize_tensor(x, interval_thresholds, n_bits, signed),
                                                      interval_thresholds).reshape([-1, 1])

            imp_losses = (interval_losses < best['loss']).reshape([-1, 1])

            best = {"param": interval_thresholds * imp_losses + best['param'] * np.logical_not(imp_losses),
                    "loss": interval_losses * imp_losses + best['loss'] * np.logical_not(imp_losses)}
    else:
        # search per-tensor
        intervals = np.linspace(start=range_bounds[0], stop=range_bounds[1], num=n_intervals, dtype=float)
        interval_losses = list(map(lambda t: loss_fn(x, quantize_tensor(x, t, n_bits, signed), t), intervals))
        best = {"param": intervals[np.argmin(interval_losses)], "loss": np.min(interval_losses)}

    return best


def search_dynamic_range(base_range: np.ndarray, x: np.ndarray, scalers: np.ndarray, loss_fn: Callable, n_bits: int,
                         per_channel: bool = False) -> Dict[str, np.ndarray]:
    """
    Searches in a set of constructed quantization ranges.

    Args:
        base_range: Base quantization range for constructing ranges candidates.
        x: Numpy array with tensor's content.
        scalers: A set of scale factor for constructing ranges candidates.
        loss_fn: Function to compute the error between the original and quantized tensors.
        n_bits: Number of bits to quantize the
        per_channel: Whether the search is done per-channel or per-tensor.

    Returns: Dictionary with best obtained quantization range and the threshold's matching loss.

    """
    if per_channel:
        # search per-channel
        ranges = np.stack([np.multiply.outer(base_range[:, 0], scalers[:, 0]),
                           np.multiply.outer(base_range[:, 1], scalers[:, 1])], axis=2)
        first_ranges_set = ranges[:, 0, :]
        best = {"param": first_ranges_set,
                "loss": _error_function_wrapper(loss_fn, x,
                                                uniform_quantize_tensor(x,
                                                                        first_ranges_set[:, 0].reshape([-1, 1]),
                                                                        first_ranges_set[:, 1].reshape([-1, 1]),
                                                                        n_bits),
                                                first_ranges_set).reshape([-1, 1])}
        for range_idx in range(1, ranges.shape[1]):  # iterate over sets of per-channel ranges
            ranges_set = ranges[:, range_idx, :]  # takes the i'th range for each channel
            ranges_losses = _error_function_wrapper(loss_fn, x,
                                                    uniform_quantize_tensor(x,
                                                                            ranges_set[:, 0].reshape([-1, 1]),
                                                                            ranges_set[:, 1].reshape([-1, 1]),
                                                                            n_bits),
                                                    ranges_set).reshape([-1, 1])
            imp_losses = ranges_losses < best['loss']
            best = {"param": ranges_set * imp_losses + best['param'] * np.logical_not(imp_losses),
                    "loss": ranges_losses * imp_losses + best['loss'] * np.logical_not(imp_losses)}
    else:
        # search per-tensor
        ranges = base_range * scalers
        ranges_losses = list(map(lambda mm: loss_fn(x, uniform_quantize_tensor(x, mm[0], mm[1], n_bits), mm), ranges))
        best = {"param": ranges[np.argmin(ranges_losses)], "loss": np.min(ranges_losses)}

    return best


def qparams_symmetric_selection_tensor_search(error_function: Callable,
                                              tensor_data: np.ndarray,
                                              tensor_max: np.ndarray,
                                              n_bits: int,
                                              per_channel: bool = False,
                                              channel_axis: int = 1,
                                              n_iter: int = SYMMETRIC_TENSOR_PER_CHANNEL_N_ITER,
                                              min_threshold=MIN_THRESHOLD) -> Any:
    """
    Search for optimal threshold (per-channel or per-tensor) for symmetric quantization of a tensor,
    using the iterative optimizer method.

    Args:
        error_function: Function to compute the error between the original and quantized tensors.
        tensor_data: Numpy array with tensor's content.
        tensor_max: The max value of the tensor.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the tensor should be quantized per-channel or per-tensor.
        channel_axis: Index of output channels dimension.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Ndarray with an optimized threshold (or set of thresholds shaped according to the channels_axis if per-channel).

    """

    signed = np.any(tensor_data < 0)  # check if tensor is singed
    output_shape = get_output_shape(tensor_data.shape, channel_axis)

    # If the threshold is computed per-channel, we rearrange the tensor such that each sub-tensor
    # is flattened, and we iterate over each one of them when searching for the threshold.
    if per_channel:
        tensor_data_r = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)
        max_tensor = np.maximum(min_threshold, tensor_max)
        res = qparams_symmetric_iterative_minimization(x0=max_tensor,
                                                       x=tensor_data_r,
                                                       loss_fn=error_function,  # gets float_tensor, fxp_tensor, threshold
                                                       n_bits=n_bits,
                                                       signed=signed,
                                                       n_intervals=SYMMETRIC_TENSOR_PER_CHANNEL_N_INTERVALS,
                                                       n_iter=SYMMETRIC_TENSOR_PER_CHANNEL_N_ITER,
                                                       dec_freq=SYMMETRIC_TENSOR_PER_CHANNEL_DEC_FREQ,
                                                       per_channel=True)
        return np.reshape(np.maximum(min_threshold, res['param']), output_shape)
    else:
        # quantize per-tensor
        res = qparams_symmetric_iterative_minimization(x0=get_init_threshold(min_threshold, tensor_max),
                                                       x=tensor_data,
                                                       loss_fn=error_function,
                                                       n_bits=n_bits,
                                                       signed=signed,
                                                       n_intervals=SYMMETRIC_TENSOR_N_INTERVALS,
                                                       n_iter=SYMMETRIC_TENSOR_N_ITER,
                                                       dec_freq=SYMMETRIC_TENSOR_DEC_FREQ,
                                                       per_channel=False)

        return max(min_threshold, res['param'])


def qparams_uniform_selection_tensor_search(error_function: Callable,
                                            tensor_data: np.ndarray,
                                            tensor_min: np.ndarray,
                                            tensor_max: np.ndarray,
                                            n_bits: int,
                                            per_channel: bool = False,
                                            channel_axis: int = 1,
                                            n_iter: int = UNIFORM_TENSOR_PER_CHANNEL_N_ITER) -> Any:
    """
    Search for optimal quantization range (per-channel or per-tensor) for uniform quantization of a tensor,
    using the iterative optimizer method and built-in scale factors
    for constructing ranges candidates during the search.

    Args:
        error_function: Function to compute the error between the original and quantized tensors.
        tensor_data: Numpy array with tensor's content.
        tensor_min: The min value of the tensor.
        tensor_max: The max value of the tensor.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the tensor should be quantized per-channel or per-tensor.
        channel_axis: Index of output channels dimension.
        n_iter: Number of searching iterations.

    Returns:
        Ndarray with an optimized range (or set of thresholds shaped according to the channels_axis if per-channel).

    """

    output_shape = get_output_shape(tensor_data.shape, channel_axis)

    alpha = np.linspace(BOTTOM_FACTOR, UPPER_FACTOR, UNIFORM_TENSOR_N_SAMPLES)
    beta = np.linspace(BOTTOM_FACTOR, UPPER_FACTOR, UNIFORM_TENSOR_N_SAMPLES)
    scalers = np.asarray(list(itertools.product(alpha, beta)))

    # If the threshold is computed per-channel, we rearrange the tensor such that each sub-tensor
    # is flattened, and we iterate over each one of them when searching for the threshold.
    if per_channel:
        if per_channel:
            tensor_data_r = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)
            tensor_min_max = np.column_stack([tensor_min.flatten(), tensor_max.flatten()])
            res = iterative_uniform_dynamic_range_search(x0=tensor_min_max,
                                                         x=tensor_data_r,
                                                         scalers=scalers,
                                                         loss_fn=error_function,
                                                         n_bits=n_bits,
                                                         n_iter=UNIFORM_TENSOR_PER_CHANNEL_N_ITER,
                                                         per_channel=True)
            return np.reshape(res['param'][:, 0], output_shape), np.reshape(res['param'][:, 1], output_shape)
    else:
        # quantize per-tensor
        pass
        res = iterative_uniform_dynamic_range_search(x0=np.array([tensor_min, tensor_max]),
                                                     x=tensor_data,
                                                     scalers=scalers,
                                                     loss_fn=error_function,
                                                     n_bits=n_bits,
                                                     n_iter=UNIFORM_TENSOR_N_ITER,
                                                     per_channel=False)
        return res['param']


def qparams_symmetric_selection_histogram_search(error_function: Callable,
                                                 tensor_max: np.ndarray,
                                                 bins: np.ndarray,
                                                 counts: np.ndarray,
                                                 n_bits: int,
                                                 n_iter: int = SYMMETRIC_HISTOGRAM_N_ITER,
                                                 min_threshold: float = MIN_THRESHOLD):
    """
    search for optimal threshold (per-channel or per-tensor) for symmetric quantization of a histogram,
    using the iterative optimizer method.

    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        tensor_max: The max value of the tensor.
        bins: Bins of the histogram to search_methods for an optimal threshold.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Optimized threshold for quantifying the histogram.

    """
    signed = np.any(bins[:-1][counts != 0] < 0)  # Whether histogram contains negative values or not.

    res = qparams_symmetric_iterative_minimization(x0=get_init_threshold(min_threshold, tensor_max),
                                                   x=bins,
                                                   loss_fn=lambda x, q_x, t:  # t is dummy, we only need it for KL error
                                                   qparams_selection_histogram_search_error_function(error_function,
                                                                                                     x,
                                                                                                     q_x,
                                                                                                     counts),
                                                   n_bits=n_bits,
                                                   signed=signed,
                                                   n_intervals=SYMMETRIC_HISTOGRAM_N_INTERVALS,
                                                   n_iter=SYMMETRIC_HISTOGRAM_N_ITER,
                                                   dec_freq=SYMMETRIC_HISTOGRAM_DEC_FREQ,
                                                   per_channel=False)
    return max(min_threshold, res['param'])


def kl_qparams_symmetric_selection_histogram_search(error_function: Callable,
                                                    tensor_max: np.ndarray,
                                                    bins: np.ndarray,
                                                    counts: np.ndarray,
                                                    n_bits: int,
                                                    n_iter: int = SYMMETRIC_HISTOGRAM_N_ITER,
                                                    min_threshold: float = MIN_THRESHOLD):
    """
    Search for optimal threshold (per-channel or per-tensor) for symmetric quantization of a histogram,
    with KL-Divergence loss function (needs a separate search function
    since the error function needs additional arguments that are constructed from the input)
    Using the iterative optimizer method for the search.

    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        tensor_max: The max value of the tensor.
        bins: Bins of the histogram to search_methods for an optimal threshold.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Optimized threshold for quantifying the histogram.

    """
    signed = np.any(bins[:-1][counts != 0] < 0)  # Whether histogram contains negative values or not.
    res = qparams_symmetric_iterative_minimization(x0=get_init_threshold(min_threshold, tensor_max),
                                                   x=bins,
                                                   loss_fn=lambda x, q_x, t:
                                                   kl_qparams_selection_histogram_search_error_function(error_function,
                                                                                                        bins,
                                                                                                        q_x,
                                                                                                        counts,
                                                                                                        min_max_range=np.array(
                                                                                                            [0,
                                                                                                             t]) if not signed else np.array(
                                                                                                            [-t, t])),
                                                   n_bits=n_bits,
                                                   signed=signed,
                                                   n_intervals=SYMMETRIC_HISTOGRAM_N_INTERVALS,
                                                   n_iter=SYMMETRIC_HISTOGRAM_N_ITER,
                                                   dec_freq=SYMMETRIC_HISTOGRAM_DEC_FREQ,
                                                   per_channel=False)
    return max(min_threshold, res['param'])


def qparams_uniform_selection_histogram_search(error_function: Callable,
                                               tensor_min_max: np.ndarray,
                                               bins: np.ndarray,
                                               counts: np.ndarray,
                                               n_bits: int,
                                               n_iter: int = UNIFORM_HISTOGRAM_N_ITER):
    """
    Search for optimal quantization range (per-channel or per-tensor) for uniform quantization of a histogram,
    using the iterative optimizer method and built-in scale factors
    for constructing ranges candidates during the search.

    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        tensor_min_max: Numpy array with tensor's min and max values.
        bins: Bins of the histogram to search_methods for an optimal threshold.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        n_iter: Number of searching iterations.

    Returns:
        Optimized range for quantifying the histogram.

    """
    alpha = np.linspace(BOTTOM_FACTOR, UPPER_FACTOR, UNIFORM_HISTOGRAM_N_SAMPLES)
    beta = np.linspace(BOTTOM_FACTOR, UPPER_FACTOR, UNIFORM_HISTOGRAM_N_SAMPLES)
    scalers = np.asarray(list(itertools.product(alpha, beta)))
    res = iterative_uniform_dynamic_range_search(x0=tensor_min_max,
                                                 x=bins,
                                                 scalers=scalers,
                                                 loss_fn=lambda x, q_x, mm:
                                                 qparams_selection_histogram_search_error_function(error_function,
                                                                                                   x,
                                                                                                   q_x,
                                                                                                   counts,
                                                                                                   min_max_range=mm),
                                                 n_bits=n_bits,
                                                 n_iter=UNIFORM_HISTOGRAM_N_ITER,
                                                 per_channel=False)
    return res['param']


def qparams_selection_histogram_search_error_function(error_function: Callable,
                                                      bins: np.ndarray,
                                                      q_bins: np.ndarray,
                                                      counts: np.ndarray,
                                                      threshold: np.ndarray = None,
                                                      min_max_range=None):
    """
    Computes the error according to the given error function, to be used in the parameters' selection process
    for quantization.
    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        bins: Bins values of the histogram.
        q_bins: Bins values of the quantized histogram.
        counts: Bins counts of the original histogram.
        threshold: Threshold bins were quantized by (used only for kl error function).
        min_max_range: quantization parameter, used in uniform parameters' selection for quantization range validation
        (not used for symmetric parameters' selection)

    Returns: the error between the original and quantized histogram.

    """
    # computes the number of elements between quantized bin values.
    q_count, _ = np.histogram(q_bins, bins=bins, weights=np.concatenate([counts.flatten(), np.asarray([0])]))
    # threshold is only used for KL error method calculations.
    # other error methods are passed with a wrapper that accepts a threshold argument but does not use it.
    error = error_function(q_bins, q_count, bins, counts, threshold, min_max_range)  # computes the error
    return error


def kl_qparams_selection_histogram_search_error_function(error_function: Callable,
                                                         bins: np.ndarray,
                                                         q_bins: np.ndarray,
                                                         counts: np.ndarray,
                                                         min_max_range: np.ndarray):
    """
    Computes the error according to the KL-divergence the distributions of the given histogram.
    The error value is used in the threshold optimization process, for symmetric quantization.
    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        bins: Bins values of the histogram.
        q_bins: Quantized bins values of the histogram.
        counts: Bins counts of the original histogram.
        min_max_range: Quantization range to quantize histogram by.

    Returns: the error between the original and quantized histogram.

    """
    # compute the number of elements between quantized bin values.
    q_count, _ = np.histogram(q_bins, bins=bins, weights=np.concatenate([counts.flatten(), np.asarray([0])]))
    error = error_function(q_bins, q_count, bins, counts, range_min=min_max_range[0], range_max=min_max_range[1])
    return error


def get_init_threshold(min_threshold: float, tensor_max: np.ndarray, per_channel: bool = False) -> np.ndarray:
    """
    Gets an initial value for the threshold optimization process.
    If per_channel then returns a vector with initial value for each channel.

    Args:
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        tensor_max: Max value of a tensor.
        per_channel: Whether the quantization should be per-channel or not.

    Returns:
        Threshold value if max value in tensor is larger than min_threshold.
    """
    if per_channel:
        init_t = tensor_max
        init_t[tensor_max < min_threshold] = min_threshold
        return init_t
    return max(min_threshold, tensor_max)


def _error_function_wrapper(error_function: Callable,
                            float_tensor: np.ndarray,
                            q_tensor: np.ndarray,
                            in_params: np.ndarray) -> np.ndarray:
    """
    Wrapper method for using the error methods in a vectorized per-channel parameters' search.

    Args:
        error_function: error_function: Function to compute the error between the original and quantized tensors.
        float_tensor: Numpy array with float tensor's content.
        q_tensor: Numpy array with quantized tensor's content.
        in_params: Quantization params the tensor is quantized by (used in specific error functions only).

    Returns: A list of error values per-channel for the quantized tensor, according to the error function.
    """
    _error_per_list = []
    for j in range(float_tensor.shape[0]):  # iterate all channels of the tensor.
        _error_per_list.append(error_function(float_tensor[j, :], q_tensor[j, :], in_params[j]))
    return np.asarray(_error_per_list)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import operator
from typing import Tuple, Any

import torch.nn.functional
from torch.nn import Conv2d, Linear, PReLU, ELU, Hardswish, Dropout, ZeroPad2d, SiLU
from torch import reshape
from torch.nn.functional import hardswish, silu, prelu, elu
from torch.nn.functional import avg_pool2d

from model_compression_toolkit import CoreConfig, FrameworkInfo
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import BaseNode, Graph
from model_compression_toolkit.core.common.graph.graph_matchers import EdgeMatcher
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common.substitutions.shift_negative_activation import apply_shift_negative_correction
from model_compression_toolkit.core.pytorch.constants import PAD, VALUE, PADDING, BIAS, USE_BIAS

"""
This substitution aims to solve an issue of activation with negative outputs where
the portion of the negative range is relatively small. In a symmetric quantization this causes 
of bit loosing as the entire negative quantization range does not contain
any values. To solve it, we shift the output of the activation by the minimal output value (quantized) such
that all values after the shifting are positive. To correct the impact of such shifting, a correction
to the next linear node is computed and added to its bias term.
If the linear node pads the input tensor with zeros, we modify the padded value as well.  
"""


def shift_negative_activation_node_matchers():
    # Match activation nodes with negative outputs.
    snc_node = NodeOperationMatcher(PReLU) | \
               NodeOperationMatcher(prelu) | \
               NodeOperationMatcher(ELU) | \
               NodeOperationMatcher(elu) | \
               NodeOperationMatcher(Hardswish) | \
               NodeOperationMatcher(hardswish) | \
               NodeOperationMatcher(SiLU) | \
               NodeOperationMatcher(silu)

    # Match linear layers where we can add a correction.
    linear_node = NodeOperationMatcher(Conv2d) | \
                  NodeOperationMatcher(Linear)

    # Match nodes that can be in between the non-linear node to the linear node,
    # and still the substitution can be applied correctly.
    bypass_node = NodeOperationMatcher(reshape) | \
                  NodeOperationMatcher(avg_pool2d) | \
                  NodeOperationMatcher(Dropout)

    # Match a pad node that can be in between the non-linear node to the linear node.
    pad_node = NodeOperationMatcher(ZeroPad2d)

    return snc_node, linear_node, bypass_node, pad_node


def create_add_node(add_value: float,
                    prev_node_name: str,
                    input_shape: tuple) -> BaseNode:
    """
    Create a new Add node, with a constant to add.
    The name of the node is determined by its previous node's name.

    Args:
        add_value: Constant to add to to the node's input tensor.
        prev_node_name: The name of the node before the Add node
        input_shape: Shape of the Add node's input tensor.

    Returns:
        New Add node.
    """

    add_node_name = prev_node_name + '_post_add'

    add_node = common.graph.functional_node.FunctionalNode(add_node_name,
                                                           {},
                                                           input_shape,
                                                           input_shape,
                                                           weights={},
                                                           quantization_attr={},
                                                           functional_op=operator.add,
                                                           op_call_kwargs={},
                                                           op_call_args=[float(add_value)],
                                                           layer_class=operator.add)
    return add_node


def create_pad_node(next_node_name: str,
                    prev_node_name: str,
                    value_to_pad: float,
                    input_shape: tuple,
                    pad_top: int,
                    pad_btm: int,
                    pad_left: int,
                    pad_right: int) -> BaseNode:
    """
    Create a pad node with a constant value to pad its input tensor.

    Args:
        next_node_name: Name of the node next to the pad node.
        prev_node_name: Name of the node previous to the pad node.
        value_to_pad: Constant to use for padding the input of the node.
        input_shape: Shape of input tensor.
        pad_top: Number of elements to pad above the tensor.
        pad_btm: Number of elements to pad below the tensor.
        pad_left: Number of elements to pad left to the tensor.
        pad_right: Number of elements to pad right to the tensor.

    Returns:
        A pad node which pads its input with a constant value.
    """

    pad_node_name = next_node_name + '_pre_pad'

    op_call_kwargs = {PAD: [pad_left, pad_right, pad_top, pad_btm],
               VALUE: float(value_to_pad)}

    padded_shape = input_shape[0]
    padded_shape[1] += pad_top + pad_btm
    padded_shape[2] += pad_left + pad_right
    pad_node = common.graph.functional_node.FunctionalNode(pad_node_name,
                                                           {},
                                                           input_shape,
                                                           tuple(padded_shape),
                                                           weights={},
                                                           quantization_attr={},
                                                           functional_op=torch.nn.functional.pad,
                                                           op_call_kwargs=op_call_kwargs,
                                                           op_call_args=[],
                                                           layer_class=torch.nn.functional.pad)
    return pad_node


def compute_op2d_padding():
    """
    Dummy function, needed for API.
    Returns:
        None
    """
    return None


def get_padding_values(op2d_node: BaseNode) -> Tuple[Any, Any]:
    """

    Args:
        op2d_node: convolution type node from which to extract the padding values.

    Returns:
        A tuple of containing the padding attribute and padding values.
    """
    padding, padding_values = None, None
    if isinstance(op2d_node.framework_attr.get(PADDING), int):
        padding = op2d_node.framework_attr.get(PADDING)
        padding_values = padding, padding, padding, padding
        op2d_node.framework_attr[PADDING] = 0
    elif isinstance(op2d_node.framework_attr.get(PADDING), tuple):
        padding = op2d_node.framework_attr.get(PADDING)
        padding_values = padding[0], padding[0], padding[1], padding[1]
        op2d_node.framework_attr[PADDING] = 0
    return padding, padding_values


def is_padding_node_and_node_has_padding(pad_node_to_consider: BaseNode,
                                         next_node: BaseNode) -> bool:
    """

    Args:
        pad_node_to_consider: Pad node between the non-linear and linear nodes to consider when
        correcting the expected shift.
        next_node: The next node after the node in check for correction.

    Returns:
        Whether a padding node exists and the next node is a linear node with padding.
    """
    padding = next_node.framework_attr.get(PADDING)
    return pad_node_to_consider is not None and (
                (isinstance(padding, int) and padding > 0) or (isinstance(padding, tuple) and sum(padding) > 0))


def pytorch_apply_shift_negative_correction(graph: Graph,
                                            core_config: CoreConfig,
                                            fw_info: FrameworkInfo) -> Graph:
    """
    Apply shift negative correction (SNC) on a graph built from a Pytorch model.

    Args:
        graph: Graph to apply SNC on.
        core_config: Quantization configuration.
        fw_info: FrameworkInfo object with information about the specific framework's module.

    Returns:
        Graph after SNC.
    """
    snc_node, linear_node, bypass_node, pad_node = shift_negative_activation_node_matchers()
    return apply_shift_negative_correction(graph,
                                           core_config,
                                           fw_info,
                                           snc_node,
                                           linear_node,
                                           bypass_node,
                                           pad_node,
                                           create_add_node,
                                           get_padding_values,
                                           create_pad_node,
                                           is_padding_node_and_node_has_padding,
                                           PADDING,
                                           BIAS,
                                           USE_BIAS
                                           )


# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import ABC, abstractmethod
from typing import Callable
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common import Graph
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.gptq.common.gptq_graph import get_compare_points
from model_compression_toolkit.core.common.model_builder_mode import ModelBuilderMode


class GPTQTrainer(ABC):
    """
    Abstract GPTQ training class for fine-tuning a quantized model
    """
    def __init__(self,
                 graph_float: Graph,
                 graph_quant: Graph,
                 gptq_config: GradientPTQConfig,
                 fw_impl: FrameworkImplementation,
                 fw_info: FrameworkInfo):
        """
        Build two models from a graph: A teacher network (float model) and a student network (quantized model).
        Use the dataset generator to pass images through the teacher and student networks to get intermediate
        layers outputs. Use the outputs to compute the observed loss and to back-propagate the error
        in the student network, to minimize it in the next similar steps.
        All parameters (such as number of iterations, optimizer, etc.) are in GradientPTQConfig.
        Args:
            graph_float: Graph to build a float networks from.
            graph_quant: Graph to build a quantized networks from.
            gptq_config: GradientPTQConfig with parameters about the tuning process.
            fw_impl: Framework implementation
            fw_info: Framework information
        """
        self.graph_float = graph_float
        self.graph_quant = graph_quant
        self.gptq_config = gptq_config
        self.fw_impl = fw_impl
        self.fw_info = fw_info

        # ----------------------------------------------
        # Build two models and create compare nodes
        # ----------------------------------------------
        self.compare_points, _, self.compare_points_mean, self.compare_points_std = get_compare_points(self.graph_float)

        self.float_model, self.float_user_info = fw_impl.model_builder(self.graph_float,
                                                                       mode=ModelBuilderMode.FLOAT,
                                                                       append2output=self.compare_points,
                                                                       fw_info=self.fw_info)

        self.fxp_model, self.gptq_user_info = self.build_gptq_model()

    @abstractmethod
    def build_gptq_model(self):
        """
        Build the GPTQ model with QuantizationWrappers
        Returns:
            Quantized graph for GPTQ fine-tuning, GPTQ graph user info
        """
        raise NotImplemented(f'{self.__class__.__name__} have to implement the '
                             f'framework\'s GPTQ model builder method.')

    @abstractmethod
    def train(self, representative_data_gen: Callable):
        """
        Train the quantized model using GPTQ training process
        Args:
            representative_data_gen: Dataset to use for inputs of the models.
        """
        raise NotImplemented(f'{self.__class__.__name__} have to implement the '
                             f'framework\'s train method.')

    @abstractmethod
    def update_graph(self) -> Graph:
        """
        Update a graph using GPTQ after minimizing the loss between the float model's output
        and the quantized model's outputs.
        Returns:
            Updated graph after GPTQ.
        """
        raise NotImplemented(f'{self.__class__.__name__} have to implement the '
                             f'framework\'s update_graph method.')


def gptq_training(graph_float: Graph,
                  graph_quant: Graph,
                  gptq_config: GradientPTQConfig,
                  representative_data_gen: Callable,
                  fw_impl: FrameworkImplementation,
                  fw_info: FrameworkInfo) -> Graph:
    """
    GPTQ training process using knowledge distillation with a teacher network (float model) and a student network (quantized model).
    Args:
        graph_float: Graph to build a float networks from.
        graph_quant: Graph to build a quantized networks from.
        gptq_config: GradientPTQConfig with parameters about the tuning process.
        representative_data_gen: Dataset to use for inputs of the models.
        fw_impl: Framework implementation
        fw_info: Framework information

    Returns:
        Quantized graph for export

    """
    # Get GPTQ object and initialize it
    gptq_trainer_obj = fw_impl.get_gptq_trainer_obj()
    gptq_trainer = gptq_trainer_obj(graph_float, graph_quant, gptq_config, fw_impl, fw_info)

    # Training process
    gptq_trainer.train(representative_data_gen)

    # Update graph
    graph_quant = gptq_trainer.update_graph()
    return graph_quant



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Callable, Tuple, Any

from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common import FrameworkInfo
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.model_builder_mode import ModelBuilderMode
from model_compression_toolkit.core.common.quantization.quantize_graph_weights import quantize_graph_weights

from model_compression_toolkit.core.common.substitutions.apply_substitutions import substitute
from model_compression_toolkit.core.common.user_info import UserInformation

from model_compression_toolkit.core.common.visualization.tensorboard_writer import TensorboardWriter


def _quantize_model(tg: Graph,
                    fw_info: FrameworkInfo,
                    fw_impl: FrameworkImplementation,
                    tb_w: TensorboardWriter) -> Tuple[Any, UserInformation]:
    """
    Quantize graph's weights, and build a quantized framework model from it.

    Args:
        tg: A prepared for quantization graph.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.).
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
        tb_w: TensorBoardWriter object to log events.

    Returns:
        Quantized model in the input framework, and information the user may need in order to use the quantized model.
    """

    quantized_tg = quantize_graph_weights(tg,
                                          fw_info=fw_info,
                                          fw_impl=fw_impl)
    if tb_w is not None:
        tb_w.add_graph(quantized_tg, 'after_quantization')
    ######################################
    # Back2Framework
    ######################################
    # Before building a quantized model, first apply some substitutions.
    quantized_tg = substitute(quantized_tg,
                              fw_impl.get_substitutions_pre_build())
    quantized_model, user_info = fw_impl.model_builder(quantized_tg,
                                                       mode=ModelBuilderMode.QUANTIZED,
                                                       fw_info=fw_info)
    return quantized_model, user_info


def export_model(tg,
                 fw_info,
                 fw_impl,
                 tb_w,
                 bit_widths_config):

    """
    A function for quantizing the graph's weights and build a quantized framework model from it.

    Args:
        tg: A prepared for quantization graph.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.).
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
        tb_w: TensorBoardWriter object to log events.
        bit_widths_config: mixed-precision bit configuration to be added to model user_info

    Returns:
        Quantized model in the input framework, and information the user may need in order to use the quantized model.
    """
    quantized_model, user_info = _quantize_model(tg,
                                                 fw_info,
                                                 fw_impl,
                                                 tb_w)
    user_info.mixed_precision_cfg = bit_widths_config

    return quantized_model, user_info



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import torch
from torch.nn import Module

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.pytorch.utils import set_model


def node_builder(n: BaseNode) -> Module:
    """
    Build a Pytorch module from a node.

    Args:
        n: Node to build its Pytorch layer

    Returns:
        Pytorch module that was built from the node.
    """

    framework_attr = copy.copy(n.framework_attr)
    node_instance = n.type(**framework_attr)
    node_instance.load_state_dict({k: torch.Tensor(v) for k, v in n.weights.items()}, strict=False)
    set_model(node_instance)
    return node_instance


# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import math
from copy import deepcopy
from typing import Any, Tuple

import numpy as np

from model_compression_toolkit.core.common.framework_info import FrameworkInfo, ChannelAxis
from model_compression_toolkit.core.common.collectors.histogram_collector import HistogramCollector
from model_compression_toolkit.core.common.collectors.mean_collector import MeanCollector
from model_compression_toolkit.core.common.collectors.min_max_per_channel_collector import MinMaxPerChannelCollector


class BaseStatsCollector(object):
    """
    Base class for statistics collection (contains multiple collectors such as mean collector,
    histogram collector, etc.).
    """

    def require_collection(self) -> bool:
        """
        Returns whether this tensor requires statistics collection or not.
        Should be implemented in extending classes.
        """
        raise Exception(f'require_collection is not implemented in {self.__class__.__name__}')

    def update_statistics(self,
                          x: Any):
        """
        Update statistics in collectors with a new tensor to consider.

        Args:
            x: Tensor.
        """
        raise Exception(f'update_statistics is not implemented in {self.__class__.__name__}')


class StatsCollector(BaseStatsCollector):
    """
    Class to wrap all statistics that are being collected for an input/output node.
    """

    def __init__(self,
                 out_channel_axis: int,
                 init_min_value: float = None,
                 init_max_value: float = None):
        """
        Instantiate three statistics collectors: histogram, mean and min/max per channel.
        Set initial min/max values if are known.

        Args:
            out_channel_axis: Index of output channels.
            init_min_value: Initial min value for min/max stored values.
            init_max_value: Initial max value for min/max stored values.
        """

        super().__init__()
        self.hc = HistogramCollector()
        self.mc = MeanCollector(axis=out_channel_axis)
        self.mpcc = MinMaxPerChannelCollector(init_min_value=init_min_value,
                                              init_max_value=init_max_value,
                                              axis=out_channel_axis)

    def update_statistics(self, x: Any):
        """
        Update statistics in all collectors with a new tensor to consider.

        Args:
            x: Tensor to consider when updating statistics.
        """

        x = standardize_tensor(x)
        self.hc.update(x)
        self.mc.update(x)
        self.mpcc.update(x)

    def get_mean(self) -> np.ndarray:
        """
        Get mean per-channel from mean collector. When its accessed from outside the tensor,
        the scale and shift come into consideration.

        Returns: Mean per-channel from mean collector.
        """

        return self.mc.state

    def get_min_max_values(self) -> Tuple[float, float]:
        """
        Get min/max from collector.
        When its accessed from outside the tensor, the scale and shift come into consideration.

        Returns: Min/max from collector.
        """

        min_value = self.mpcc.min
        max_value = self.mpcc.max

        if not self.mpcc.ignore_init_values:
            if is_number(self.mpcc.init_min_value):
                min_value = self.mpcc.init_min_value

            if is_number(self.mpcc.init_max_value):
                max_value = self.mpcc.init_max_value

        return float(min_value), float(max_value)

    def __repr__(self) -> str:
        """
        Display Tensor with its current and initial mix/max values.
        Returns: String to display.
        """

        return f"<Min:{self.mpcc.min}," \
               f" Max:{self.mpcc.max} " \
               f"with init values " \
               f"Min:{self.mpcc.init_min_value}, " \
               f"Max:{self.mpcc.init_max_value}>"

    def require_collection(self) -> bool:
        """
        Returns: True since tensor requires statistics collection.
        """

        return True


class NoStatsCollector(BaseStatsCollector):
    """
    Class that inherits from base tensor.
    Indicating that for a point in a graph we should not gather statistics.
    """

    def __init__(self):
        super().__init__()

    def update_statistics(self,
                          x: Any):
        """
        Do nothing in BaseTensor method since we are not collecting statistics here.

        Args:
            x: Tensor.
        """

        pass  # pragma: no cover

    def __repr__(self):
        """
        Returns: Display object as "No Quantization".
        """

        return "No Stats Collector"

    def require_collection(self):
        """
        Returns: False since NoTensor does not statistics collection.
        """

        return False


def is_number(num: Any) -> bool:
    """
    Check if a variable is an actual number (and not None or math.inf).
    Args:
        num: Variable to check.

    Returns:
        Whether the vriable is a number or not.
    """
    if num is not None and \
            not math.isinf(num):
        return True
    return False


def standardize_tensor(x: Any) -> np.ndarray:
    """
    Standardize tensors that goes through the collectors before using them.
    Convert them to numpy arrays of float data type.

    Args:
        x: Tensor to standardize.

    Returns:
        Same tensor as numpy ndarray of float data type.
    """
    x = x.astype(np.float)
    if not isinstance(x, np.ndarray) and len(x.shape) == 0:
        x = np.asarray(x)
        x = x.reshape([1])
    return x


def shift_statistics(collector: BaseStatsCollector,
                     shift_value: np.ndarray) -> BaseStatsCollector:
    """
    Shift all statistics in collectors of a statistics collector by a
    value (or a value per-channel).

    Args:
        collector: Statistics collector to shift its collectors.
        shift_value: Value to shift all statistics by.

    Returns:
        New copy of the collector with shifted statistics.

    """

    shifted_collector = deepcopy(collector)
    if isinstance(collector, StatsCollector):
        shifted_collector.mpcc.shift(shift_value)
        shifted_collector.mc.shift(shift_value)
        if shifted_collector.require_collection():
            shifted_collector.hc.shift(shift_value)

    return shifted_collector


def scale_statistics(collector: BaseStatsCollector,
                     scale_value: np.ndarray) -> BaseStatsCollector:
    """
    Scale all statistics in collectors of a statistics collector
    by a factor (or a factor per-channel).

    Args:
        collector: Statistics collector to shift its collectors.
        scale_value: Value to shift all statistics by.

    Returns:
        New copy of the collector with scaled statistics.

    """

    scaled_collector = deepcopy(collector)
    if isinstance(collector, StatsCollector):
        scaled_collector.mpcc.scale(scale_value)
        scaled_collector.mc.scale(scale_value)
        if scaled_collector.require_collection():
            scaled_collector.hc.scale(scale_value)

    return scaled_collector



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf

from model_compression_toolkit.core.common.model_builder_mode import ModelBuilderMode
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from typing import Tuple, Any
from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core import common
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.keras.default_framework_info import DEFAULT_KERAS_INFO
from model_compression_toolkit.gptq.keras.quantizer.config_factory import quantization_config_builder_gptq
from model_compression_toolkit.core.keras.back2framework.model_builder import get_node_name_from_layer, \
    is_layer_fake_quant
from model_compression_toolkit.core.keras.back2framework.model_builder import model_builder as core_model_builder


def model_builder(graph: common.Graph,
                  gptq_config: GradientPTQConfig,
                  mode: ModelBuilderMode = ModelBuilderMode.QUANTIZED,
                  append2output=None,
                  fw_info: FrameworkInfo = DEFAULT_KERAS_INFO) -> Tuple[tf.keras.models.Model, Any]:
    """
    Build a Keras model for GPTQ from a graph representing the model.
    The model is built by converting the graph nodes to Keras layers and applying them sequentially to get the model
    output tensors. The output tensors list and an input tensors list, then use to build the model.
    After the model is built in Keras, it is cloned to add quantization wrappers for GPTQ fine-tuning

    Args:
        graph: Graph to build its corresponding Keras model.
        gptq_config: GPTQ Configuration class.
        mode: Building mode. Read ModelBuilderMode description for more info.
        append2output: List of nodes or OutTensor objects. In float building mode,
        when the list contains nodes, all output tensors of all nodes are set as the model outputs.
        fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
        This is for passing the kernel attributes to the QuanteWrappers.

    Returns:
        A tuple of the model and an UserInformation object.
    """
    if gptq_config is None:
        Logger.exception("Building a model in GPTQ requires a GPTQ configuration as input")

    model, graph_user_info = core_model_builder(graph, mode, append2output, fw_info, return_float_outputs=True)

    def _quantize(layer):
        nodes = graph.find_node_by_name(get_node_name_from_layer(layer))
        if len(nodes) == 1:
            node = nodes[0]
            return QuantizeWrapper(layer, quantization_config_builder_gptq(node, fw_info, gptq_config))
        elif is_layer_fake_quant(layer):
            return layer
        else:
            raise Exception(
                f"Mismatch between keras model and graph can't find node named: {get_node_name_from_layer(layer)}")

    # clone each layer in the model and apply _quantize to the layer.
    model = tf.keras.models.clone_model(model, input_tensors=None, clone_function=_quantize)

    return model, graph_user_info



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import tensorflow as tf

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if tf.__version__ < "2.6":
    from tensorflow.python.keras.layers import Layer
else:
    from keras.engine.base_layer import Layer

from typing import Any, Dict, List
from tensorflow.python.util.object_identity import Reference as TFReference
from model_compression_toolkit.core.common.constants import EPS
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import BaseNode, Graph
from model_compression_toolkit.core.common.graph.edge import EDGE_SINK_INDEX
from model_compression_toolkit.core.keras.back2framework.instance_builder import OperationHandler


def build_input_tensors_list(node: BaseNode,
                             graph: Graph,
                             node_to_output_tensors_dict: Dict[BaseNode, List[TFReference]]) -> List[List[TFReference]]:
    """
    Given a node, build a list of input tensors the node gets. The list is built
    based on the node's incoming edges and previous nodes' output tensors.

    Args:
        node: Node to build its input tensors list.
        graph: Graph the node is in.
        node_to_output_tensors_dict: A dictionary from a node to its output tensors.

    Returns:
        A list of the node's input tensors.
    """

    input_tensors = []
    # Go over a sorted list of the node's incoming edges, and for each source node get its output tensors.
    # Append them in a result list.
    for ie in graph.incoming_edges(node, sort_by_attr=EDGE_SINK_INDEX):
        _input_tensors = [node_to_output_tensors_dict[ie.source_node][ie.source_index]]
        input_tensors.append(_input_tensors)
    return input_tensors


def run_operation(n: BaseNode,
                  input_tensors: List[List[TFReference]],
                  op_func: Layer,
                  input_nodes_to_input_tensors: Dict[BaseNode, Any]) -> List[TFReference]:
    """
    Applying the layer (op_func) to the input tensors (input_tensors).

    Args:
        n: The corresponding node of the layer it runs.
        input_tensors: List of references to Keras tensors that are the layer's inputs.
        op_func: Layer to apply to the input tensors.
        input_nodes_to_input_tensors: A dictionary from a node to its input tensors.

    Returns:
        A list of references to Keras tensors. The layer's output tensors after applying the
        layer to the input tensors.
    """

    if len(input_tensors) == 0:  # Placeholder handling
        out_tensors_of_n = input_nodes_to_input_tensors[n]
    else:
        input_tensors = [tensor for tensor_list in input_tensors for tensor in tensor_list]  # flat list of lists
        # Build a functional node using its args
        if isinstance(n, FunctionalNode):
            if n.inputs_as_list:  # If the first argument should be a list of tensors:
                out_tensors_of_n = op_func(input_tensors, *n.op_call_args, **n.op_call_kwargs)
            else:  # If the input tensors should not be a list but iterated:
                out_tensors_of_n = op_func(*input_tensors, *n.op_call_args, **n.op_call_kwargs)
        else:
            # If operator expects a single input tensor, it cannot be a list as it should have a dtype field.
            if len(input_tensors) == 1:
                input_tensors = input_tensors[0]
            out_tensors_of_n = op_func(input_tensors)

    return out_tensors_of_n


def keras_model_grad(graph_float: common.Graph,
                     model_input_tensors: Dict[BaseNode, np.ndarray],
                     interest_points: List[BaseNode],
                     output_list: List[BaseNode],
                     all_outputs_indices: List[int],
                     alpha: float = 0.3) -> List[float]:
    """
    Computes the gradients of a Keras model's outputs with respect to the feature maps of the set of given
    interest points. It then uses the gradients to compute the hessian trace for each interest point and normalized the
    values, to be used as weights for weighted average in mixed-precision distance metric computation.

    Args:
        graph_float: Graph to build its corresponding Keras model.
        model_input_tensors: A mapping between model input nodes to an input batch.
        interest_points: List of nodes which we want to get their feature map as output, to calculate distance metric.
        output_list: List of nodes that considered as model's output for the purpose of gradients computation.
        all_outputs_indices: Indices of the model outputs and outputs replacements (if exists),
            in a topological sorted interest points list.
        alpha: A tuning parameter to allow calibration between the contribution of the output feature maps returned
            weights and the other feature maps weights (since the gradient of the output layers does not provide a
            compatible weight for the distance metric computation).

    Returns: A list of normalized gradients to be considered as the relevancy that each interest
    point's output has on the model's output.
    """

    node_to_output_tensors_dict = dict()

    # Build an OperationHandler to handle conversions from graph nodes to Keras operators.
    oh = OperationHandler(graph_float)

    input_nodes_to_input_tensors = {inode: tf.convert_to_tensor(model_input_tensors[inode]) for
                                    inode in graph_float.get_inputs()}  # Cast numpy array to tf.Tensor

    # for interest point p in interest_points:
    interest_points_tensors = []
    output_tensors = []
    with tf.GradientTape(persistent=True) as g:
        # Build a dictionary from node to its output tensors, by applying the layers sequentially.
        for n in oh.node_sort:
            op_func = oh.get_node_op_function(n)  # Get node operation function

            input_tensors = build_input_tensors_list(n,
                                                     graph_float,
                                                     node_to_output_tensors_dict)  # Fetch Node inputs
            out_tensors_of_n = run_operation(n,  # Run node operation and fetch outputs
                                             input_tensors,
                                             op_func,
                                             input_nodes_to_input_tensors)

            # Gradients can be computed only on float32 tensors
            out_tensors_of_n = tf.dtypes.cast(out_tensors_of_n, tf.float32)
            if n in interest_points:
                # Recording the relevant feature maps onto the gradient tape
                g.watch(out_tensors_of_n)
                interest_points_tensors.append(out_tensors_of_n)
            if n in output_list:
                output_tensors.append(out_tensors_of_n)

            if isinstance(out_tensors_of_n, list):
                node_to_output_tensors_dict.update({n: out_tensors_of_n})
            else:
                node_to_output_tensors_dict.update({n: [out_tensors_of_n]})

        # Get a reduced loss value for derivatives computation
        output_loss = 0
        for output in output_tensors:
            output = tf.reshape(output, shape=(output.shape[0], -1))
            output_loss += tf.reduce_mean(tf.reduce_sum(output, axis=-1))

    ###########################################
    # Compute Gradients
    ##########################################
    ipt_grad_score = []

    for ipt in interest_points_tensors:
        grad_ipt = g.gradient(output_loss, ipt, unconnected_gradients=tf.UnconnectedGradients.ZERO)

        r_grad_ipt = tf.reshape(grad_ipt, shape=[grad_ipt.shape[0], -1])
        hessian_trace_aprrox = tf.reduce_mean(tf.reduce_sum(tf.pow(r_grad_ipt, 2.0), axis=-1))
        ipt_grad_score.append(hessian_trace_aprrox)

    # Output layers or layers that come after the model's considered output layers,
    # are assigned with a constant normalized value,
    # according to the given alpha variable and the number of such layers.
    # Other layers returned weights are normalized by dividing the hessian value by the sum of all other values.
    sum_without_outputs = sum([ipt_grad_score[i] for i in range(len(ipt_grad_score)) if i not in all_outputs_indices])
    normalized_grads_weights = [get_normalized_weight(grad, i, sum_without_outputs, all_outputs_indices, alpha)
                                for i, grad in enumerate(ipt_grad_score)]

    return normalized_grads_weights


def get_normalized_weight(grad: float,
                          i: int,
                          sum_without_outputs: float,
                          all_outputs_indices: List[int],
                          alpha: float) -> float:
    """
    Normalizes the node's gradient value. If it is an output or output replacement node than the normalized value is
    a constant, otherwise, it is normalized by dividing with the sum of all gradient values.

    Args:
        grad: The gradient value.
        i: The index of the node in the sorted interest points list.
        sum_without_outputs: The sum of all gradients of nodes that are not considered outputs.
        all_outputs_indices: A list of indices of all nodes that consider outputs.
        alpha: A multiplication factor.

    Returns: A normalized gradient value.

    """

    if i in all_outputs_indices:
        return alpha / len(all_outputs_indices)
    else:
        return ((1 - alpha) * grad / (sum_without_outputs + EPS)).numpy()



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from functools import partial
from typing import List

import numpy as np

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.core.common import Graph
from model_compression_toolkit.core.common.constants import BITS_TO_BYTES


def weights_size_kpi(mp_cfg: List[int], graph: Graph, fw_info: FrameworkInfo) -> np.ndarray:
    """
    Computes a KPIs vector with the respective weights' memory size for each weight configurable node,
    according to the given mixed-precision configuration.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize).

    Returns: A vector of node's weights memory sizes.
    Note that the vector is not necessarily of the same length as the given config.

    """
    weights_memory = []

    # Go over all nodes that should be taken into consideration when computing the weights KPI.
    mp_nodes = graph.get_configurable_sorted_nodes_names()
    for n in graph.get_sorted_weights_configurable_nodes():
        node_idx = mp_nodes.index(n.name)
        node_qc = n.candidates_quantization_cfg[mp_cfg[node_idx]]
        node_nbits = node_qc.weights_quantization_cfg.weights_n_bits

        node_num_weights_params = 0
        for attr in fw_info.get_kernel_op_attributes(n.type):
            if attr is not None:
                node_num_weights_params += n.get_weights_by_keys(attr).flatten().shape[0]

        node_weights_memory_in_bytes = node_num_weights_params * node_nbits / BITS_TO_BYTES
        weights_memory.append(node_weights_memory_in_bytes)

    return np.array(weights_memory)


def activation_output_size_kpi(mp_cfg: List[int], graph: Graph, fw_info: FrameworkInfo) -> np.ndarray:
    """
    Computes a KPIs vector with the respective output memory size for each activation configurable node,
    according to the given mixed-precision configuration.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize)
            (not used in this method).

    Returns: A vector of node's activation memory sizes.
    Note that the vector is not necessarily of the same length as the given config.

    """

    activation_memory = []

    # Go over all nodes that should be taken into consideration when computing the weights KPI.
    mp_nodes = graph.get_configurable_sorted_nodes_names()
    for n in graph.get_sorted_activation_configurable_nodes():
        node_idx = mp_nodes.index(n.name)
        node_qc = n.candidates_quantization_cfg[mp_cfg[node_idx]]
        node_nbits = node_qc.activation_quantization_cfg.activation_n_bits

        node_output_size = n.get_total_output_params()
        node_activation_memory_in_bytes = node_output_size * node_nbits / BITS_TO_BYTES
        activation_memory.append(node_activation_memory_in_bytes)

    return np.array(activation_memory)


def total_weights_activation_kpi(mp_cfg: List[int], graph: Graph, fw_info: FrameworkInfo) -> np.ndarray:
    """
    Computes KPIs tensor with the respective weights size and output memory size for each activation configurable node,
    according to the given mixed-precision configuration.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize)
            (not used in this method).

    Returns: A 2D tensor of nodes' weights memory sizes and activation output memory size.
    Note that the vector is not necessarily of the same length as the given config.

    """
    weights_activation_memory = []

    # Go over all nodes that should be taken into consideration when computing the weights or
    # activation KPI (all configurable nodes).
    for node_idx, n in enumerate(graph.get_configurable_sorted_nodes()):
        node_qc = n.candidates_quantization_cfg[mp_cfg[node_idx]]
        node_weights_nbits = node_qc.weights_quantization_cfg.weights_n_bits
        node_activation_nbits = node_qc.activation_quantization_cfg.activation_n_bits

        # Compute node's weights memory (if no weights to quantize then set to 0)
        node_num_weights_params = 0
        for attr in fw_info.get_kernel_op_attributes(n.type):
            if attr is not None:
                node_num_weights_params += n.get_weights_by_keys(attr).flatten().shape[0]
        node_weights_memory_in_bytes = node_num_weights_params * node_weights_nbits / BITS_TO_BYTES

        # Compute node's activation memory (if node's activation are not being quantized then set to 0)
        node_activation_memory_in_bytes = 0
        if n.is_activation_quantization_enabled():
            node_output_size = n.get_total_output_params()
            node_activation_memory_in_bytes = node_output_size * node_activation_nbits / BITS_TO_BYTES

        weights_activation_memory.append(np.array([node_weights_memory_in_bytes, node_activation_memory_in_bytes]))

    return np.array(weights_activation_memory)


class MpKpiMetric(Enum):
    """
    Defines kpi computation functions that can be used to compute KPI for a given target for a given mp config.
    The enum values can be used to call a function on a set of arguments.

     WEIGHTS_SIZE - applies the weights_size_kpi function

     ACTIVATION_OUTPUT_SIZE - applies the activation_output_size_kpi function

     TOTAL_WEIGHTS_ACTIVATION_SIZE - applies the total_weights_activation_kpi function

    """

    WEIGHTS_SIZE = partial(weights_size_kpi)
    ACTIVATION_OUTPUT_SIZE = partial(activation_output_size_kpi)
    TOTAL_WEIGHTS_ACTIVATION_SIZE = partial(total_weights_activation_kpi)

    def __call__(self, *args):
        return self.value(*args)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
from model_compression_toolkit.core.keras.constants import KERNEL


def get_kernel(weights_list: list) -> tf.Tensor:
    """
    This function a list of weights and return the kernel
    Args:
        weights_list:  A list of Tensors

    Returns: The kernel tensor.

    """
    for w in weights_list:
        if KERNEL in w.name:
            return w
    raise Exception("Can't find kernel variable")


def threshold_reshape(threshold_tensor: tf.Tensor, input_w: tf.Tensor, in_quantization_axis: int) -> tf.Tensor:
    """
    This function take a threshold tensor and re-aline it to the weight tensor channel axis.
    Args:
        threshold_tensor: A tensor of threshold
        input_w: A weight tensor
        in_quantization_axis: A int value that represent the channel axis.

    Returns: A reshape tensor of threshold.

    """
    input_shape = input_w.shape
    n_axis = len(input_shape)
    quantization_axis = n_axis + in_quantization_axis if in_quantization_axis < 0 else in_quantization_axis
    reshape_shape = [-1 if i == quantization_axis else 1 for i in range(n_axis)]
    ptq_threshold_tensor = tf.reshape(threshold_tensor, reshape_shape)
    return ptq_threshold_tensor



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import logging
import os
from datetime import datetime
from os import path
from pathlib import Path

LOGGER_NAME = 'Constrained Model Optimization'


class Logger:
    # Logger has levels of verbosity.
    log_level_translate = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL
    }
    LOG_PATH = None

    @staticmethod
    def __check_path_create_dir(log_path: str):
        """
        Create a path if not exist. Otherwise, do nothing.
        Args:
            log_path: Path to create or verify that exists.

        """

        if not path.exists(log_path):
            Path(log_path).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def set_logger_level(log_level=logging.INFO):
        """
        Set log level to determine the logger verbosity.
        Args:
            log_level: Level of verbosity to set for the logger.

        """

        logger = Logger.get_logger()
        logger.setLevel(log_level)

    @staticmethod
    def get_logger():
        """
        Returns: An instance of the logger.
        """
        return logging.getLogger(LOGGER_NAME)

    @staticmethod
    def set_log_file(log_folder: str = None):
        """
        Setting the logger log file path. The method gets the folder for the log file.
        In that folder, it creates a log file according to the timestamp.
        Args:
            log_folder: Folder path to hold the log file.

        """

        logger = Logger.get_logger()

        ts = datetime.now(tz=None).strftime("%d%m%Y_%H%M%S")

        if log_folder is None:
            Logger.LOG_PATH = os.path.join(os.environ.get('LOG_PATH', os.getcwd()), f"logs_{ts}")
        else:
            Logger.LOG_PATH = os.path.join(log_folder, f"logs_{ts}")
        log_name = os.path.join(Logger.LOG_PATH, f'mct_log.log')

        Logger.__check_path_create_dir(Logger.LOG_PATH)

        fh = logging.FileHandler(log_name)
        fh.setLevel(logging.DEBUG)
        logger.addHandler(fh)

        print(f'log file is in {log_name}')

    ########################################
    # Delegating methods to wrapped logger
    ########################################

    @staticmethod
    def critical(msg: str):
        """
        Log a message at 'critical' severity and raise an exception.
        Args:
            msg: Message to log.

        """
        Logger.get_logger().critical(msg)
        raise Exception(msg)

    @staticmethod
    def exception(msg: str):
        """
        Log a message at 'exception' severity and raise an exception.
        Args:
            msg: Message to log.

        """
        Logger.get_logger().exception(msg)
        raise Exception(msg)

    @staticmethod
    def debug(msg: str):
        """
        Log a message at 'debug' severity.

        Args:
            msg: Message to log.

        """
        Logger.get_logger().debug(msg)

    @staticmethod
    def info(msg: str):
        """
        Log a message at 'info' severity.

        Args:
            msg: Message to log.

        """
        print(msg)
        Logger.get_logger().info(msg)

    @staticmethod
    def warning(msg: str):
        """
        Log a message at 'warning' severity.

        Args:
            msg: Message to log.

        """
        print(msg)
        Logger.get_logger().warning(msg)

    @staticmethod
    def error(msg: str):
        """
        Log a message at 'error' severity and raise an exception.

        Args:
            msg: Message to log.

        """
        Logger.get_logger().error(msg)
        raise Exception(msg)


def set_log_folder(folder: str, level: int = logging.INFO):
    """
    Set a directory path for saving a log file.

    Args:
        folder: Folder path to save the log file.
        level: Level of verbosity to set to the logger.

    """
    Logger.set_log_file(folder)
    Logger.set_logger_level(level)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from model_compression_toolkit.core.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.core.common.quantization.debug_config import DebugConfig
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import MixedPrecisionQuantizationConfigV2


class CoreConfig:
    """
    A class to hold the configurations classes of the MCT-core.
    """
    def __init__(self, n_iter: int = 500,
                 quantization_config: QuantizationConfig = QuantizationConfig(),
                 mixed_precision_config: MixedPrecisionQuantizationConfigV2 = None,
                 debug_config: DebugConfig = DebugConfig()
                 ):
        """

        Args:
            n_iter (int): Number of calibration iterations to run.
            quantization_config (QuantizationConfig): Config for quantization.
            mixed_precision_config (MixedPrecisionQuantizationConfigV2): Config for mixed precision quantization (optional, default=None).
            debug_config (DebugConfig): Config for debugging and editing the network quantization process.
        """
        self.n_iter = n_iter
        self.quantization_config = quantization_config
        self.mixed_precision_config = mixed_precision_config
        self.debug_config = debug_config

    @property
    def mixed_precision_enable(self):
        return self.mixed_precision_config is not None



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common import FrameworkInfo
from model_compression_toolkit.core.common.graph.base_graph import Graph

from model_compression_toolkit.core.common.visualization.tensorboard_writer import TensorboardWriter
from model_compression_toolkit.core.common.bias_correction.apply_bias_correction_to_graph import \
    apply_bias_correction_to_graph


def ptq_runner(tg: Graph,
               fw_info: FrameworkInfo,
               fw_impl: FrameworkImplementation,
               tb_w: TensorboardWriter) -> Graph:
    """
    Quantize a graph that has final weights candidates quantization configurations.

    Args:
        tg: Graph to apply GPTQ and to quantize.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.)
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
        tb_w: A TensorBoardWriter object initialized with the logger dir path if it was set, or None otherwise.

    Returns:
        A graph after bias correction

    """
    #############################################
    # Apply Bias Correction
    #############################################
    tg_bias = apply_bias_correction_to_graph(tg,
                                             fw_impl=fw_impl)
    if tb_w is not None:
        tb_w.add_graph(tg_bias, 'after_bias_correction')

    return tg_bias



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import tensorflow as tf
from tensorflow.python.framework.tensor_shape import TensorShape
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from tensorflow_model_optimization.python.core.quantization.keras.quantizers import Quantizer
from typing import Dict, Any, List, Callable

from model_compression_toolkit.core.common.quantization.candidate_node_quantization_config import \
    CandidateNodeQuantizationConfig
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeActivationQuantizationConfig


class SelectiveActivationQuantizer(Quantizer):
    """
    Quantizer that can use different quantized weights on-the-fly.
    The general idea behind this kind of quantizer is that it gets the float tensor to quantize
    when initialize, it quantizes the float tensor in different bitwidths, and every time it need to return a
    quantized version of the float weight, it returns only one quantized weight according to an "active"
    index - the index of a candidate weight quantization configuration from a list of candidates that was passed
    to the SelectiveWeightsQuantizer when it was initialized.
    The "active" index can be configured as part of the SelectiveWeightsQuantizer's API, so a different quantized
    weight can be returned in another time.
    """

    def __init__(self,
                 node_q_cfg: List[CandidateNodeQuantizationConfig],
                 max_candidate_idx: int):
        """
        Init a selective quantizer.

        Args:
            node_q_cfg: Quantization configuration candidate of the node that generated the layer that will
                use this quantizer.
            max_candidate_idx: Index of the node's candidate that has the maximal bitwidth (must exist absolute max).
        """
        self.node_q_cfg = node_q_cfg
        self.active_quantization_config_index = max_candidate_idx  # initialize with first config as default
        self.activation_quantizers = []
        self._store_activation_quantizers()

    def _get_qc_quantizer(self, index: int) -> NodeActivationQuantizationConfig:
        """
        Quantize the quantizer float weight using a candidate quantization configuration.

        Args:
            index: Index of the candidate to use for the quantization.

        Returns:
            Quantized weight.
        """
        qc = self.node_q_cfg[index].activation_quantization_cfg
        return qc

    def _store_activation_quantizers(self):
        """
        Go over all candidates configurations, quantize the quantizer float weight according to each one
        of them, and store the quantized weights in a list quantized_weights the quantizer holds.
        """
        for i in range(len(self.node_q_cfg)):
            q_activation = self._get_qc_quantizer(i)
            self.activation_quantizers.append(q_activation.quantize_node_output)

    def build(self,
              tensor_shape: TensorShape,
              name: str,
              layer: QuantizeWrapper) -> Dict[str, tf.Variable]:
        """
        The build method has to be implemented as part of the Keras framework,
        but there is no need to use it here as we do not train any new variable.
        Hence, it returns an empty dictionary.
        """

        return {}

    def __call__(self,
                 inputs: tf.Tensor,
                 training: bool,
                 weights: Dict[str, tf.Variable],
                 **kwargs: Dict[str, Any]) -> np.ndarray:
        """
        Method to return the quantized weight. This method is called
        when the framework needs to quantize a float weight, and is expected to return the quantized
        weight. Since we already quantized the weight in all possible bitwidths, we do not
        quantize it again, and simply return the quantized weight according to the current
        active_quantization_config_index.

        Returns:
            Quantized weight, that was quantized using number of bits that is in a
            specific quantization configuration candidate (the candidate's index is the
            index that is in active_quantization_config_index the quantizer holds).
        """
        return self.activation_quantizers[self.active_quantization_config_index](inputs)

    def set_active_quantization_config_index(self, index: int):
        """
        Set an index to use for the quantized weight the quantizer returns
        when requested.

        Args:
            index: Index of a candidate quantization configuration to use its quantized
            version of the float weight.
        """
        assert index < len(
            self.node_q_cfg), f'Quantizer has {len(self.node_q_cfg)} ' \
                                      f'possible nbits. Can not set ' \
                                      f'index {index}'
        self.active_quantization_config_index = index

    def get_active_quantization_config_index(self) -> int:
        """
        Returns: The index the quantizer uses when is selects which quantized weight
        to use when asked to (in __call__).
        """
        return self.active_quantization_config_index

    def get_config(self) -> Dict[str, Any]:
        """
        Returns: Configuration of TrainableQuantizer.
        """

        return {
            'node_q_cfg': self.node_q_cfg,
        }

    def __eq__(self, other: Any) -> bool:
        """
        Check if equals to another object.

        Args:
            other: Other object to compare.

        Returns:
            Whether they are equal or not.
        """
        if not isinstance(other, SelectiveActivationQuantizer):
            return False

        return self.node_q_cfg == other.node_q_cfg

    def __ne__(self, other: Any) -> bool:
        """
        Check if not equals to another object.

        Args:
            other: Other object to compare.

        Returns:
            Whether they are differ or not.
        """
        return not self.__eq__(other)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
from typing import Callable, Any, List

from model_compression_toolkit import FrameworkInfo, MixedPrecisionQuantizationConfigV2
from model_compression_toolkit.core.common import Graph, BaseNode
from model_compression_toolkit.core.common.model_builder_mode import ModelBuilderMode
from model_compression_toolkit.core.common import Logger


class SensitivityEvaluation:
    """
    Class to wrap and manage the computation on distance metric for Mixed-Precision quantization search.
    It provides a function that evaluates the sensitivity of a bit-width configuration for the MP model.
    """

    def __init__(self,
                 graph: Graph,
                 quant_config: MixedPrecisionQuantizationConfigV2,
                 representative_data_gen: Callable,
                 fw_info: FrameworkInfo,
                 fw_impl: Any,
                 set_layer_to_bitwidth: Callable,
                 get_quant_node_name: Callable):
        """
        Initiates all relevant objects to manage a sensitivity evaluation for MP search.
        Create an object that allows to compute the sensitivity metric of an MP model (the sensitivity
        is computed based on the similarity of the interest points' outputs between the MP model
        and the float model).
        First, we initiate a SensitivityEvaluationManager that handles the components which are necessary for
        evaluating the sensitivity. It initializes an MP model (a model where layers that can be configured in
        different bit-widths) and a baseline model (a float model).
        Then, and based on the outputs of these two models (for some batches from the representative_data_gen),
        we build a function to measure the sensitivity of a change in a bit-width of a model's layer.

        Args:
            graph: Graph to search for its MP configuration.
            fw_info: FrameworkInfo object about the specific framework
                (e.g., attributes of different layers' weights to quantize).
            quant_config: MP Quantization configuration for how the graph should be quantized.
            representative_data_gen: Dataset used for getting batches for inference.
            fw_impl: FrameworkImplementation object with a specific framework methods implementation.
            set_layer_to_bitwidth: A fw-dependent function that allows to configure a configurable MP model
                    with a specific bit-width configuration.
            get_quant_node_name: A fw-dependent function that takes a node's name and outputs the node's name in a
                quantized model (according to the fw conventions).
        """
        self.graph = graph
        self.quant_config = quant_config
        self.representative_data_gen = representative_data_gen
        self.fw_info = fw_info
        self.fw_impl = fw_impl
        self.set_layer_to_bitwidth = set_layer_to_bitwidth
        self.get_quant_node_name = get_quant_node_name

        # Get interest points for distance measurement and a list of sorted configurable nodes names
        self.sorted_configurable_nodes_names = graph.get_configurable_sorted_nodes_names()
        self.interest_points = get_mp_interest_points(graph,
                                                      fw_impl.count_node_for_mixed_precision_interest_points,
                                                      quant_config.num_interest_points_factor)

        self.outputs_replacement_nodes = None
        self.output_nodes_indices = None
        if self.quant_config.use_grad_based_weights is True:
            # Getting output replacement (if needed) - if a model's output layer is not compatible for the task of
            # gradients computation then we find a predecessor layer which is compatible,
            # add it to the set of interest points and use it for the gradients' computation.
            # Note that we need to modify the set of interest points before building the models,
            # therefore, it is separated from the part where we compute the actual gradient weights.
            self.outputs_replacement_nodes = get_output_replacement_nodes(graph, fw_impl)
            self.output_nodes_indices = self._update_ips_with_outputs_replacements()

        # Build a mixed-precision model which can be configured to use different bitwidth in different layers.
        # And a baseline model.
        self.baseline_model, self.model_mp = self._build_models()

        # Build images batches for inference comparison
        self.images_batches = self._get_images_batches(quant_config.num_of_images)

        # Get baseline model inference on all samples
        self.baseline_tensors_list = []  # setting from outside scope

        # Casting images tensors to the framework tensor type.
        self.images_batches = list(map(lambda in_arr: self.fw_impl.to_tensor(in_arr), self.images_batches))

        # Initiating baseline_tensors_list since it is not initiated in SensitivityEvaluationManager init.
        self._init_baseline_tensors_list()

        # Computing gradient-based weights for weighted average distance metric computation (only if requested),
        # and assigning distance_weighting method accordingly.
        self.interest_points_gradients = None
        if self.quant_config.use_grad_based_weights is True:
            assert self.outputs_replacement_nodes is not None and self.output_nodes_indices is not None, \
                f"{self.outputs_replacement_nodes} and {self.output_nodes_indices} " \
                f"should've been assigned before computing the gradient-based weights."

            self.interest_points_gradients = self._compute_gradient_based_weights()
            self.quant_config.distance_weighting_method = lambda d: self.interest_points_gradients

    def compute_metric(self,
                       mp_model_configuration: List[int],
                       node_idx: List[int] = None,
                       baseline_mp_configuration: List[int] = None) -> float:
        """
        Compute the sensitivity metric of the MP model for a given configuration (the sensitivity
        is computed based on the similarity of the interest points' outputs between the MP model
        and the float model).

        Args:
            mp_model_configuration: Bitwidth configuration to use to configure the MP model.
            node_idx: A list of nodes' indices to configure (instead of using the entire mp_model_configuration).
            baseline_mp_configuration: A mixed-precision configuration to set the model back to after modifying it to
                compute the metric for the given configuration.

        Returns:
            The sensitivity metric of the MP model for a given configuration.
        """

        # Configure MP model with the given configuration.
        self._configure_bitwidths_model(self.model_mp,
                                        self.sorted_configurable_nodes_names,
                                        mp_model_configuration,
                                        node_idx)

        # Compute the distance matrix
        distance_matrix = self._build_distance_metrix()

        # Configure MP model back to the same configuration as the baseline model if baseline provided
        if baseline_mp_configuration is not None:
            self._configure_bitwidths_model(self.model_mp,
                                            self.sorted_configurable_nodes_names,
                                            baseline_mp_configuration,
                                            node_idx)

        return self._compute_mp_distance_measure(distance_matrix, self.quant_config.distance_weighting_method)

    def _init_baseline_tensors_list(self):
        """
        Evaluates the baseline model on all images and saves the obtained lists of tensors in a list for later use.
        Initiates a class variable self.baseline_tensors_list
        """
        self.baseline_tensors_list = [self._tensors_as_list(self.fw_impl.to_numpy(self.baseline_model(images)))
                                      for images in self.images_batches]

    def _build_models(self) -> Any:
        """
        Builds two models - an MP model with configurable layers and a baseline, float model.

        Returns: A tuple with two models built from the given graph: a baseline model (with baseline configuration) and
            an MP model (which can be configured for a specific bitwidth configuration).
            Note that the type of the returned models is dependent on the used framework (TF/Pytorch).
        """
        # Build a mixed-precision model which can be configured to use different bitwidth in different layers.
        model_mp, _ = self.fw_impl.model_builder(self.graph,
                                                 mode=ModelBuilderMode.MIXEDPRECISION,
                                                 append2output=self.interest_points,
                                                 fw_info=self.fw_info)

        # Build a baseline model.
        baseline_model, _ = self.fw_impl.model_builder(self.graph,
                                                       mode=ModelBuilderMode.FLOAT,
                                                       append2output=self.interest_points)

        return baseline_model, model_mp

    def _compute_gradient_based_weights(self) -> np.ndarray:
        """
        Computes the gradient-based weights using the framework's model_grad method per batch of images.

        Returns: A vector of weights, one for each interest point,
        to be used for the distance metric weighted average computation.
        """

        grad_per_batch = []
        for images in self.images_batches:
            batch_ip_gradients = self.fw_impl.model_grad(self.graph,
                                                         {inode: images[0] for inode in self.graph.get_inputs()},
                                                         self.interest_points,
                                                         self.outputs_replacement_nodes,
                                                         self.output_nodes_indices,
                                                         self.quant_config.output_grad_factor)

            grad_per_batch.append(batch_ip_gradients)
        return np.mean(grad_per_batch, axis=0)

    def _configure_bitwidths_model(self,
                                   model_mp: Any,
                                   sorted_configurable_nodes_names: List[str],
                                   mp_model_configuration: List[int],
                                   node_idx: List[int]):
        """
        Configure a dynamic model (namely, model with layers that their weights and activation
        bit-width can be configured) using an MP model configuration mp_model_configuration.

        Args:
            model_mp: Dynamic model to configure.
            sorted_configurable_nodes_names: List of configurable nodes names sorted topology.
            mp_model_configuration: Configuration of bit-width indices to set to the model.
            node_idx: List of nodes' indices to configure (the rest layers are configured as the baseline model).
        """

        # Configure model
        # Note: Not all nodes in the graph are included in the MP model that is returned by the model builder.
        # Thus, the last configurable layer must be included in the interest points for evaluating the metric,
        # otherwise, not all configurable nodes will be considered throughout the MP optimization search (since
        # they will not affect the metric value).
        model_mp_layers_names = self.fw_impl.get_model_layers_names(model_mp)
        if node_idx is not None:  # configure specific layers in the mp model
            for node_idx_to_configure in node_idx:
                node_name = self.get_quant_node_name(sorted_configurable_nodes_names[node_idx_to_configure])
                if node_name in model_mp_layers_names:
                    current_layer = self.fw_impl.get_model_layer_by_name(model_mp, node_name)
                    self.set_layer_to_bitwidth(current_layer, mp_model_configuration[node_idx_to_configure])
                else:
                    raise Exception("The last configurable node is not included in the list of interest points for"
                                    "sensitivity evaluation metric for the mixed-precision search.")

        else:  # use the entire mp_model_configuration to configure the model
            for node_idx_to_configure, bitwidth_idx in enumerate(mp_model_configuration):
                node_name = self.get_quant_node_name(sorted_configurable_nodes_names[node_idx_to_configure])
                if node_name in model_mp_layers_names:
                    current_layer = self.fw_impl.get_model_layer_by_name(model_mp, node_name)
                    self.set_layer_to_bitwidth(current_layer, mp_model_configuration[node_idx_to_configure])
                else:
                    raise Exception("The last configurable node is not included in the list of interest points for"
                                    "sensitivity evaluation metric for the mixed-precision search.")

    def _compute_distance_matrix(self,
                                 baseline_tensors: List[Any],
                                 mp_tensors: List[Any]):
        """
        Compute the distance between the MP model's outputs and the baseline model's outputs
        for each image in the batch that was inferred.
        Args:
            baseline_tensors: Baseline model's output tensors.
            mp_tensors: MP model's output tensors.
        Returns:
            A distance matrix that maps each node's index to the distance between this node's output
             and the baseline model's output for all images that were inferred.
        """

        assert len(baseline_tensors) == len(self.interest_points)
        num_interest_points = len(baseline_tensors)
        num_samples = len(baseline_tensors[0])
        distance_matrix = np.ndarray((num_interest_points, num_samples))

        for i in range(num_interest_points):
            point_distance_fn = \
                self.fw_impl.get_node_distance_fn(layer_class=self.interest_points[i].layer_class,
                                                  framework_attrs=self.interest_points[i].framework_attr,
                                                  compute_distance_fn=self.quant_config.compute_distance_fn)
            for j in range(num_samples):
                distance_matrix[i, j] = point_distance_fn(baseline_tensors[i][j], mp_tensors[i][j])

        return distance_matrix

    def _build_distance_metrix(self):
        """
        Builds a matrix that contains the distances between the baseline and MP models for each interest point.
        Returns: A distance matrix.
        """
        # List of distance matrices. We create a distance matrix for each sample from the representative_data_gen
        # and merge all of them eventually.
        distance_matrices = []

        # Compute the distance matrix for num_of_images images.
        for images, baseline_tensors in zip(self.images_batches, self.baseline_tensors_list):
            # when using model.predict(), it does not use the QuantizeWrapper functionality
            mp_tensors = self.model_mp(images)
            mp_tensors = self._tensors_as_list(self.fw_impl.to_numpy(mp_tensors))

            # Build distance matrix: similarity between the baseline model to the float model
            # in every interest point for every image in the batch.
            distance_matrices.append(self._compute_distance_matrix(baseline_tensors, mp_tensors))

        # Merge all distance matrices into a single distance matrix.
        distance_matrix = np.concatenate(distance_matrices, axis=1)

        # Assert we used a correct number of images for computing the distance matrix
        assert distance_matrix.shape[1] == self.quant_config.num_of_images
        return distance_matrix

    @staticmethod
    def _compute_mp_distance_measure(distance_matrix: np.ndarray, metrics_weights_fn: Callable) -> float:
        """
        Computes the final distance value out of a distance matrix.

        Args:
            distance_matrix: A matrix that contains the distances between the baseline and MP models
                for each interest point.
            metrics_weights_fn:

        Returns: Distance value.
        """
        # Compute the distance between the baseline model's outputs and the MP model's outputs.
        # The distance is the mean of distances over all images in the batch that was inferred.
        mean_distance_per_layer = distance_matrix.mean(axis=1)
        # Use weights such that every layer's distance is weighted differently (possibly).
        return np.average(mean_distance_per_layer, weights=metrics_weights_fn(distance_matrix))

    def _get_images_batches(self, num_of_images: int) -> List[Any]:
        """
        Construct batches of image samples for inference.

        Args:
            num_of_images: Num of total images for evaluation.

        Returns: A list of images batches (lists of images)
        """
        # First, select images to use for all measurements.
        samples_count = 0  # Number of images we used so far to compute the distance matrix.
        images_batches = []
        while samples_count < num_of_images:
            # Get a batch of images to infer in both models.
            inference_batch_input = self.representative_data_gen()
            batch_size = inference_batch_input[0].shape[0]

            # If we sampled more images than we should use in the distance matrix,
            # we take only a subset of these images and use only them for computing the distance matrix.
            if batch_size > num_of_images - samples_count:
                inference_batch_input = [x[:num_of_images - samples_count] for x in inference_batch_input]
                assert num_of_images - samples_count == inference_batch_input[0].shape[0]
                batch_size = num_of_images - samples_count

            images_batches.append(inference_batch_input)
            samples_count += batch_size
        return images_batches

    def _update_ips_with_outputs_replacements(self):
        """
        Updates the list of interest points with the set of pre-calculated replacement outputs.
        Also, returns the indices of all output nodes (original, replacements and nodes in between them) in a
        topological sorted interest points list (for later use in gradients computation and normalization).

        Returns: A list of indices of the output nodes in the sorted interest points list.

        """

        assert self.outputs_replacement_nodes is not None, \
            "Trying to update interest points list with new output nodes but outputs_replacement_nodes list is None."

        replacement_outputs_to_ip = [r_node for r_node in self.outputs_replacement_nodes if
                                     r_node not in self.interest_points]
        updated_interest_points = self.interest_points + replacement_outputs_to_ip

        # Re-sort interest points in a topological order according to the graph's sort
        self.interest_points = [n for n in self.graph.get_topo_sorted_nodes() if n in updated_interest_points]

        output_indices = [self.interest_points.index(n.node) for n in self.graph.get_outputs()]
        replacement_indices = [self.interest_points.index(n) for n in self.outputs_replacement_nodes]
        return list(set(output_indices + replacement_indices))

    @staticmethod
    def _tensors_as_list(tensors: Any) -> List[Any]:
        """
        Create a list of tensors if they are not in a list already.
        This functions helpful when the graph has only one node (so the model's output is a Tensor and not a list of
        Tensors).
        Args:
            tensors: Tensors to return as a list.
        Returns:
            List of tensors.
        """
        if not isinstance(tensors, list):
            return [tensors]
        return tensors


def get_mp_interest_points(graph: Graph,
                           interest_points_classifier: Callable,
                           num_ip_factor: float) -> List[BaseNode]:
    """
    Gets a list of interest points for the mixed precision metric computation.
    The list is constructed from a filtered set of the convolutions nodes in the graph.

    Args:
        graph: Graph to search for its MP configuration.
        interest_points_classifier: A function that indicates whether a given node in considered as a potential
            interest point for mp metric computation purposes.
        num_ip_factor: Percentage out of the total set of interest points that we want to actually use.

    Returns: A list of interest points (nodes in the graph).

    """
    sorted_nodes = graph.get_topo_sorted_nodes()
    ip_nodes = list(filter(lambda n: interest_points_classifier(n), sorted_nodes))

    interest_points_nodes = bound_num_interest_points(ip_nodes, num_ip_factor)

    # We add output layers of the model to interest points
    # in order to consider the model's output in the distance metric computation (and also to make sure
    # all configurable layers are included in the configured mp model for metric computation purposes)
    output_nodes = [n.node for n in graph.get_outputs() if n.node not in interest_points_nodes]
    interest_points = interest_points_nodes + output_nodes

    return interest_points


def get_output_replacement_nodes(graph: Graph,
                                 fw_impl: Any) -> List[BaseNode]:
    """
    If a model's output node is not compatible for the task of gradients computation we need to find a predecessor
    node in the model's graph representation which is compatible and add it to the set of interest points and use it
    for the gradients' computation. This method searches for this predecessor node for each output of the model.

    Args:
        graph: Graph to search for replacement output nodes.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    Returns: A list of output replacement nodes.

    """
    replacement_outputs = []
    for n in graph.get_outputs():
        prev_node = n.node
        while not fw_impl.is_node_compatible_for_mp_metric_outputs(prev_node):
            prev_node = graph.get_prev_nodes(n.node)
            assert len(prev_node) == 1, "A none MP compatible output node has multiple inputs, " \
                                        "which is incompatible for metric computation."
            prev_node = prev_node[0]
        replacement_outputs.append(prev_node)
    return replacement_outputs


def bound_num_interest_points(sorted_ip_list: List[BaseNode], num_ip_factor: float) -> List[BaseNode]:
    """
    Filters the list of interest points and returns a shorter list with number of interest points smaller than some
    default threshold.

    Args:
        sorted_ip_list: List of nodes which are considered as interest points for the metric computation.
        num_ip_factor: Percentage out of the total set of interest points that we want to actually use.

    Returns: A new list of interest points (list of nodes).

    """
    if num_ip_factor < 1.0:
        num_interest_points = int(num_ip_factor * len(sorted_ip_list))
        Logger.info(f'Using {num_interest_points} for mixed-precision metric evaluation out of total '
                    f'{len(sorted_ip_list)} potential interest points.')
        # Take num_interest_points evenly spaced interest points from the original list
        indices = np.round(np.linspace(0, len(sorted_ip_list) - 1, num_interest_points)).astype(int)
        return [sorted_ip_list[i] for i in indices]

    return sorted_ip_list



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, List

import torch
from networkx import topological_sort

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import BaseNode, Graph
from model_compression_toolkit.core.common.constants import EPS
from model_compression_toolkit.core.common.graph.edge import EDGE_SINK_INDEX
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.core.pytorch.back2framework.instance_builder import node_builder
from model_compression_toolkit.core.pytorch.reader.graph_builders import DummyPlaceHolder


def build_input_tensors_list(node: BaseNode,
                             graph: Graph,
                             inputs: Dict[BaseNode, List],
                             node_to_output_tensors_dict: Dict[BaseNode, List]) -> List[List]:
    """
    Given a node, build a list of input tensors the node gets. The list is built
    based on the node's incoming edges and previous nodes' output tensors.
    Args:
        node: Node to build its input tensors list.
        graph: Graph the node is in.
        inputs: list of input tensors to model
        node_to_output_tensors_dict: A dictionary from a node to its output tensors.
    Returns:
        A list of the node's input tensors.
    """
    if node.type == DummyPlaceHolder:
        input_tensors = [inputs[node]]
    else:
        input_tensors = []
        # Go over a sorted list of the node's incoming edges, and for each source node get its output tensors.
        # Append them in a result list.
        for ie in graph.incoming_edges(node, sort_by_attr=EDGE_SINK_INDEX):
            _input_tensors = node_to_output_tensors_dict[ie.source_node]
            input_tensors.append(_input_tensors)
        input_tensors = [tensor for tensor_list in input_tensors for tensor in tensor_list]  # flat list of lists
    return input_tensors


def run_operation(n: BaseNode,
                  input_tensors: List,
                  op_func: Any) -> List[torch.Tensor]:
    """
    Applying the layer (op_func) to the input tensors (input_tensors).
    If quantized is set to True, and the layer's corresponding node (n) has quantization
    attributes, an additional fake-quantization node is built and appended to the layer.
    Args:
        n: The corresponding node of the layer it runs.
        input_tensors: List of Pytorch tensors that are the layer's inputs.
        op_func: Module/functional to apply to the input tensors.
    Returns:
        A list of Pytorch tensors. The Module/functional output tensors after applying the
        Module/functional to the input tensors.
    """

    op_call_args = n.op_call_args if isinstance(n, FunctionalNode) else []
    functional_kwargs = n.op_call_kwargs if isinstance(n, FunctionalNode) else {}
    if isinstance(n, FunctionalNode) and n.inputs_as_list:
        out_tensors_of_n = op_func(input_tensors, *op_call_args, **functional_kwargs)
    else:
        out_tensors_of_n = op_func(*input_tensors + op_call_args, **functional_kwargs)

    return out_tensors_of_n


def generate_outputs(
        out_nodes: List[BaseNode],
        node_to_output_tensors_dict: dict):
    """
    Args:
        out_nodes: List of output nodes.
        node_to_output_tensors_dict: A dictionary from a node to its output tensors.
    Returns: List of output tensor/s for the model
    """
    output = []
    for n in out_nodes:
        out_tensors_of_n = node_to_output_tensors_dict.get(n)
        if len(out_tensors_of_n) > 1:
            output.append(out_tensors_of_n)
        else:
            output += out_tensors_of_n
    return output


class PytorchModelGradients(torch.nn.Module):
    """
    A Pytorch Module class for producing differentiable outputs from a given model's graph representation.
    """
    def __init__(self,
                 graph_float: common.Graph,
                 interest_points: List[BaseNode]):
        """
        Construct a PytorchModelGradients model.

        Args:
            graph_float: Model's Graph representation to evaluate the outputs according to.
            interest_points: List of nodes in the graph which we want to produce outputs for.
        """

        super(PytorchModelGradients, self).__init__()
        self.graph_float = graph_float
        self.node_sort = list(topological_sort(graph_float))
        self.interest_points = interest_points

        for n in self.node_sort:
            if not isinstance(n, FunctionalNode):
                self.add_module(n.name, node_builder(n))

    def forward(self,
                *args: Any) -> Any:
        """
        Args:
            args: argument input tensors to model, which is a mappings between an input node and its input tensor.

        Returns:
            A list of output tensors for each of the model's pre-defined interest points.
        """

        input_node_to_input_tensor = args[0]
        node_to_output_tensors_dict = dict()
        for n in self.node_sort:
            input_tensors = build_input_tensors_list(n,
                                                     self.graph_float,
                                                     input_node_to_input_tensor,
                                                     node_to_output_tensors_dict)

            op_func = n.type if isinstance(n, FunctionalNode) else getattr(self, n.name)
            out_tensors_of_n = run_operation(n,  # Run node operation and fetch outputs
                                             input_tensors,
                                             op_func=op_func)

            if isinstance(out_tensors_of_n, list):
                output_t = []
                for t in out_tensors_of_n:
                    if n in self.interest_points and t.requires_grad:
                        t.retain_grad()
                    output_t.append(t)
                node_to_output_tensors_dict.update({n: output_t})
            else:
                assert isinstance(out_tensors_of_n, torch.Tensor)
                if n in self.interest_points and out_tensors_of_n.requires_grad:
                    out_tensors_of_n.retain_grad()
                node_to_output_tensors_dict.update({n: [out_tensors_of_n]})

        outputs = generate_outputs(self.interest_points,
                                   node_to_output_tensors_dict)

        return outputs


def pytorch_model_grad(graph_float: common.Graph,
                       model_input_tensors: Dict[BaseNode, torch.Tensor],
                       interest_points: List[BaseNode],
                       all_outputs_indices: List[int],
                       alpha: float = 0.3) -> List[float]:
    """
    Computes the gradients of a Pytorch model's outputs with respect to the feature maps of the set of given
    interest points. It then uses the gradients to compute the hessian trace for each interest point and normalized the
    values, to be used as weights for weighted average in mixed-precision distance metric computation.

    Args:
        graph_float: Graph to build its corresponding Pytorch model.
        model_input_tensors: A mapping between model input nodes to an input batch torch Tensor.
        interest_points: List of nodes which we want to get their feature map as output, to calculate distance metric.
        all_outputs_indices: Indices of the model outputs and outputs replacements (if exists),
            in a topological sorted interest points list.
        alpha: A tuning parameter to allow calibration between the contribution of the output feature maps returned
            weights and the other feature maps weights (since the gradient of the output layers does not provide a
            compatible weight for the distance metric computation).

    Returns: A list of normalized gradients to be considered as the relevancy that each interest
    point's output has on the model's output.
    """

    for n, input_tensor in model_input_tensors.items():
        input_tensor.requires_grad_()

    model_grads_net = PytorchModelGradients(graph_float=graph_float,
                                            interest_points=interest_points)

    output_tensors = model_grads_net(model_input_tensors)

    ############################################
    # Compute Gradients
    ############################################
    device = output_tensors[0].device
    output_tensors_for_loss = [output_tensors[i] for i in all_outputs_indices]
    output_loss = torch.tensor([0.0], requires_grad=True, device=device)
    for output in output_tensors_for_loss:
        output = torch.reshape(output, shape=(output.shape[0], -1))
        output_loss = torch.add(output_loss, torch.mean(torch.sum(output, dim=-1)))

    output_loss.backward()

    ipt_gradients = [torch.Tensor([0.0]) if t.grad is None else t.grad for t in output_tensors]
    r_ipt_gradients = [torch.reshape(t, shape=(t.shape[0], -1)) for t in ipt_gradients]
    hessian_trace_aprrox = [torch.mean(torch.sum(torch.pow(ipt_grad, 2.0), dim=-1)) for ipt_grad in r_ipt_gradients]

    # Output layers or layers that come after the model's considered output layers,
    # are assigned with a constant normalized value,
    # according to the given alpha variable and the number of such layers.
    # Other layers returned weights are normalized by dividing the hessian value by the sum of all other values.
    hessians_without_outputs = [hessian_trace_aprrox[i] for i in range(len(hessian_trace_aprrox))
                                if i not in all_outputs_indices]
    sum_without_outputs = sum(hessians_without_outputs)
    normalized_grads_weights = [get_normalized_weight(grad,
                                                      i,
                                                      sum_without_outputs,
                                                      all_outputs_indices,
                                                      alpha)
                                for i, grad in enumerate(hessian_trace_aprrox)]

    return normalized_grads_weights


def get_normalized_weight(grad: torch.Tensor,
                          i: int,
                          sum_without_outputs: float,
                          all_outputs_indices: List[int],
                          alpha: float) -> float:
    """
    Normalizes the node's gradient value. If it is an output or output replacement node than the normalized value is
    a constant, otherwise, it is normalized by dividing with the sum of all gradient values.

    Args:
        grad: The gradient value tensor.
        i: The index of the node in the sorted interest points list.
        sum_without_outputs: The sum of all gradients of nodes that are not considered outputs.
        all_outputs_indices: A list of indices of all nodes that consider outputs.
        alpha: A multiplication factor.

    Returns: A normalized gradient value.

    """

    if i in all_outputs_indices:
        return alpha / len(all_outputs_indices)
    else:
        return ((1 - alpha) * grad / (sum_without_outputs + EPS)).item()



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.core.common.target_platform import TargetPlatformCapabilities

from model_compression_toolkit.core.tpc_models.default_tpc.target_platform_capabilities import \
    tpc_dict as default_tpc_dict
from model_compression_toolkit.core.tpc_models.tflite_tpc.target_platform_capabilities import \
    tpc_dict as tflite_tpc_dict
from model_compression_toolkit.core.tpc_models.qnnpack_tpc.target_platform_capabilities import \
    tpc_dict as qnnpack_tpc_dict
from model_compression_toolkit.core.keras.constants import DEFAULT_TP_MODEL, TFLITE_TP_MODEL, QNNPACK_TP_MODEL
from model_compression_toolkit.core.common.constants import LATEST

tpc_dict = {DEFAULT_TP_MODEL: default_tpc_dict,
            TFLITE_TP_MODEL: tflite_tpc_dict,
            QNNPACK_TP_MODEL: qnnpack_tpc_dict}


def get_target_platform_capabilities(fw_name: str,
                                     target_platform_name: str,
                                     target_platform_version: str = None) -> TargetPlatformCapabilities:
    """
    Get a TargetPlatformCapabilities by the target platform model name and the framework name.
    For now, it supports frameworks 'tensorflow' and 'pytorch'. For both of them
    the target platform model can be 'default','tflite', or 'qnnpack'.

    Args:
        fw_name: Framework name of the TargetPlatformCapabilities.
        target_platform_name: Target platform model name the model will use for inference.
        target_platform_version: Target platform capabilities version.
    Returns:
        A TargetPlatformCapabilities object that models the hardware and attaches
        a framework information to it.
    """
    assert target_platform_name in tpc_dict, f'Target platform {target_platform_name} is not defined!'
    fw_tpc = tpc_dict.get(target_platform_name)
    assert fw_name in fw_tpc, f'Framework {fw_name} is not supported in {target_platform_name}. Please make sure the relevant ' \
                              f'packages are installed when using MCT for optimizing a {fw_name} model. ' \
                              f'For Tensorflow, please install tensorflow and tensorflow-model-optimization. ' \
                              f'For PyTorch, please install torch.'
    tpc_versions = fw_tpc.get(fw_name)
    if target_platform_version is None:
        target_platform_version = LATEST
    else:
        assert target_platform_version in tpc_versions, f'TPC version {target_platform_version} is not supported for framework {fw_name}'
    return tpc_versions.get(target_platform_version)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.base_node import BaseNode


class SoftmaxShift(common.BaseSubstitution):
    """
    Shift layer by const before Softmax
    """

    def __init__(self,
                 nodes: List[BaseNode],
                 bias_str: str):
        """
        Matches: (Softmax)

        Args:
            nodes: Nodes matcher for Linear/Dense - softmax type nodes.
            bias_str: The framework specific attribute name of the convolution layer's bias.
        """
        super().__init__(matcher_instance=nodes)
        self.bias_str = bias_str

    def substitute(self,
                   graph: Graph,
                   nodes: List[BaseNode]) -> Graph:
        """
        Shift the layer before Softmax activation.

        Args:
            graph: Graph we apply the substitution on.
            nodes: nodes that match the pattern in the substitution init.

        Returns:
            Graph after applying the substitution.
        """

        first_node = nodes[0]

        if first_node.is_activation_quantization_enabled():
            tensor_stat = graph.get_out_stats_collector(first_node)
            if isinstance(tensor_stat, common.StatsCollector):
                max_value = tensor_stat.mpcc.max
                min_value = tensor_stat.mpcc.min
                shift_value = -1 * (max_value + min_value) / 2

                if first_node.get_weights_by_keys(self.bias_str) is not None:
                    b1 = first_node.get_weights_by_keys(self.bias_str)
                else:
                    b1 = 0.0
                b1_fixed = b1 + shift_value
                first_node.set_weights_by_keys(self.bias_str, b1_fixed)
                graph.shift_stats_collector(first_node, shift_value)

        return graph



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Callable

import tensorflow as tf
from model_compression_toolkit.gptq.keras.quantizer.configs.weight_quantizer_gptq_config import WeightQuantizeConfig
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper


class SAM:
    """
    This class implements Sharpness-Aware Minimization for Efficiently Improving Generalization (https://arxiv.org/abs/2010.01412)
    """

    def __init__(self, model2quantized,
                 gradient_step: Callable,
                 trainable_variables: List[tf.Tensor],
                 rho: float = 0.01,
                 eps: float = 1e-12):
        """
        The init function of Sharpness-Aware Minimization  gradient computation class.
        Args:
            model2quantized: Input quantized module
            gradient_step: A function that returns a list of gradients tensors
            trainable_variables: List of trainable variables
            rho: A floating point number that set the region of smoothness
            eps: A floating point number for numeric stability
        """
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"
        self.rho = rho
        self.eps = eps
        self.gradient_step = gradient_step

        self.trainable_variables = trainable_variables
        self.m_var = len(self.trainable_variables)
        self.model2quantized = model2quantized
        self.e_ws = []

    def _enable_update_step_param(self):
        """
        This function enables the parameter update (update iteration index and gumbel random variable)
        Returns: None

        """
        for layer in self.model2quantized.layers:
            if isinstance(layer, QuantizeWrapper) and isinstance(
                    layer.quantize_config, WeightQuantizeConfig):
                layer.quantize_config.enable_update()

    def _disable_update_step_param(self):
        """
        This function disables the parameter update (update iteration index and gumbel random variable)
        Returns: None

        """
        for layer in self.model2quantized.layers:
            if isinstance(layer, QuantizeWrapper) and isinstance(
                    layer.quantize_config, WeightQuantizeConfig):
                layer.quantize_config.disable_update()

    def _update_w_location(self, gradients: List[tf.Tensor]):
        """
        This function updates the weights position to the highest point
        Args:
            gradients: A list of gradients tensors

        Returns: None

        """
        self.e_ws.clear()
        grad_norm = tf.linalg.global_norm(gradients)
        ew_multiplier = self.rho / (grad_norm + self.eps)
        for i in range(self.m_var):
            e_w = tf.math.multiply(gradients[i], ew_multiplier)
            self.trainable_variables[i].assign_add(e_w)
            self.e_ws.append(e_w)

    def _restore_w_location(self):
        """
        Restore weights to the original position
        Returns: None

        """
        for i in range(self.m_var):
            self.trainable_variables[i].assign_add(-self.e_ws[i])

    def compute_gradients(self, *arg, **kwargs) -> (tf.Tensor, List[tf.Tensor]):
        """
        This function compute the gradients of SAM optimizer
        Args:
            *arg: args to pass to the gradient step functions
            **kwargs: kwargs to pass to the gradient step functions

        Returns: A tensor of the loss value and a list of gradients tensors

        """
        self._enable_update_step_param()
        loss, grad = self.gradient_step(*arg, **kwargs)
        self._update_w_location(grad)
        self._disable_update_step_param()
        loss, grad = self.gradient_step(*arg, **kwargs)
        self._restore_w_location()
        return loss, grad



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from abc import abstractmethod
from typing import Any


class BaseMatcher(object):
    """
    Base class for matchers. Matcher has the ability to check if an input object
    satisfies a condition the matcher checks.
    """

    @abstractmethod
    def apply(self, input_object: Any) -> bool:
        """
        Check if input_object satisfies a condition the matcher checks.
        Args:
            input_object: Object to check.

        Returns:
            True if input_object satisfies a condition the matcher checks.
            Otherwise, return nothing.
        """
        pass  # pragma: no cover

    def __and__(self, other: Any):
        """
        Return a matcher to check the logic AND of two BaseMatchers on an object.
        """
        raise NotImplemented

    def __or__(self, other: Any):
        """
        Return a matcher to check the logic OR of BaseMatchers on an object.
        """
        raise NotImplemented

    def logic_not(self):
        """
        Return a matcher to check the logic NOT of the BaseMatcher on an object.
        """
        raise NotImplemented

    def logic_and(self, other: Any):
        """
        Return a matcher to check the logic AND of two BaseMatchers on an object.
        """
        return self & other

    def logic_or(self, other: Any):
        """
        Return a matcher to check the logic OR of two BaseMatchers on an object.
        """
        return self | other
